from __future__ import annotations

import re
import typing
from collections.abc import Callable
from collections.abc import Iterable
from collections.abc import Sequence
from dataclasses import dataclass
from typing import Annotated

from docnote import Note

if typing.TYPE_CHECKING:
    from templatey.parser import LiteralTemplateString
    from templatey.parser import TemplateInstanceContentRef
    from templatey.parser import TemplateInstanceDataRef
    from templatey.parser import TemplateInstanceVariableRef


@dataclass(slots=True, frozen=True)
class SegmentModifier:
    """Segment modifiers are a particularly powerful tool for reducing
    the verbosity of the actual template text. They allow you to modify
    literal string segments that match a particular pattern, replacing
    that pattern with a different segment sequence.

    For example, you could have a modifier that replaces every newline
    in the segment with a call to an ``add_indentation`` environment
    function:

    Note that if you want access to any part of the matched pattern,
    you must include it in a capture group:
    """
    pattern: re.Pattern
    modifier: Callable[
        [SegmentModifierMatch],
        Sequence[
            EnvFuncInvocationRef
            | TemplateInstanceContentRef
            | TemplateInstanceVariableRef
            | str]]

    def apply_and_flatten(
            self,
            literal_template_string: LiteralTemplateString
            ) -> tuple[
                bool,
                list[
                    str
                    | EnvFuncInvocationRef
                    | TemplateInstanceContentRef
                    | TemplateInstanceVariableRef]]:
        """This takes a literal template string segment and applies all
        modifiers. This gets it ready to be converted into the actual
        ParsedTemplateResource parts we need, but doesn't yet perform
        the final conversion.

        Returns a tuple of [had_matches, segments_after_modification]
        """
        after_modification: list[
            str
            | EnvFuncInvocationRef
            | TemplateInstanceContentRef
            | TemplateInstanceVariableRef] = []
        had_matches = False
        splits = self.pattern.split(literal_template_string)
        for split_segment_or_modification in SegmentModifierMatch.from_splits(
            self.pattern,
            splits
        ):
            if isinstance(split_segment_or_modification, str):
                # Here's where we filter out any empty strings. This also helps
                # us re-normalize the output from re.split, so that we recover
                # from the land of "let's use an empty string as a placeholder
                # for anything we did implicitly"
                if split_segment_or_modification:
                    after_modification.append(split_segment_or_modification)
            else:
                had_matches = True
                after_modification.extend(
                    self.modifier(split_segment_or_modification))

        return had_matches, after_modification


@dataclass(slots=True, kw_only=True)
class SegmentModifierMatch:
    """
    """
    captures: Annotated[
        list[str],
        Note(
            '''The captures in a segment modifier match correspond to
            the capture groups in the ``SegmentModifier`` pattern that
            resulted in the match. If the pattern had zero match groups,
            it will be an empty list. If it had one match group, it will
            be a single-item list. Etc.

            The ordering will be the same as the ordering of the groups
            in the original pattern. Named groups are not supported.
            ''')]

    @classmethod
    def from_splits(
            cls,
            src_pattern: re.Pattern,
            splits: list[str]
            ) -> Iterable[SegmentModifierMatch | str]:
        # Note: the +1 is because we need to account for the literal strings
        # in the pattern; otherwise the mod math doesn't work out! This also
        # has the added benefit of meaning we don't need to change behavior
        # between both cases.
        split_type_count = src_pattern.groups + 1

        current_captures = []
        for index, post_split_segment in enumerate(splits):
            # This is always literal text, even if it's an empty string.
            if not index % split_type_count:
                # Note that the zeroth and last index are always literal text,
                # and are never preceeded or followed by a capture. However,
                # we still need to yield the preceeding match when we reach
                # the end.
                if 0 < index:
                    yield cls(captures=current_captures)
                    current_captures = []

                # Note: we're saving the "cull empty strings" bit for later;
                # it makes the logic much cleaner to do one at a time
                yield post_split_segment

            else:
                current_captures.append(post_split_segment)


@dataclass(slots=True, init=False)
class EnvFuncInvocationRef:
    """Used to indicate that a segment modification needs to invoke an
    environment function. Instantiate these like partials:

    > Invocation example
        def my_env_func(foo: str, *, bar: int) -> list[str]:
            ...

        EnvFuncInvocationRef('my_env_func', 'foo', bar=3)
        EnvFuncInvocationRef(
            'my_env_func',
            # TemplateInstanceContentRef, TemplateInstanceDataRef, and
            # TemplateInstanceVariableRef are all supported, including nested
            # within containers
            TemplateInstanceVariableRef('foo'),
            bar=TemplateInstanceDataRef('bar'))

    Note that unfortunately, until/unless higher-kinded type support is
    added to python, these won't be able to type-check correctly.
    """
    name: str
    call_args: tuple[
        object
        | TemplateInstanceContentRef
        | TemplateInstanceDataRef
        | TemplateInstanceVariableRef, ...]
    call_kwargs: dict[
        str,
        object
        | TemplateInstanceContentRef
        | TemplateInstanceDataRef
        | TemplateInstanceVariableRef]

    def __init__(
            self,
            name: str,
            /,
            *call_args:
                object
                | TemplateInstanceContentRef
                | TemplateInstanceDataRef
                | TemplateInstanceVariableRef,
            **call_kwargs:
                object
                | TemplateInstanceContentRef
                | TemplateInstanceDataRef
                | TemplateInstanceVariableRef):
        self.name = name
        self.call_args = call_args
        self.call_kwargs = call_kwargs



from __future__ import annotations

import typing
from dataclasses import dataclass
from dataclasses import field
from dataclasses import fields

from templatey._provenance import Provenance
from templatey._types import TemplateClass
from templatey._types import TemplateInstanceID
from templatey._types import TemplateParamsInstance

if typing.TYPE_CHECKING:
    from templatey._fields import NormalizedFieldset
    from templatey._slot_tree import PrerenderTreeNode
    from templatey._slot_tree import SlotTreeNode
    from templatey.environments import AsyncTemplateLoader
    from templatey.environments import SyncTemplateLoader
    from templatey.templates import ParseConfig
    from templatey.templates import RenderConfig

type GroupedTemplateInvocations = dict[TemplateClass, list[Provenance]]
type TemplateLookupByID = dict[TemplateInstanceID, TemplateParamsInstance]


@dataclass(slots=True, kw_only=True)
class TemplateSignature:
    """Signature objects are created immediately upon template
    definition time, and are populated with information on the template
    as it becomes available. Any object decorated with ``@template``
    will have a signature, but not all items in the signature will be
    available at all points of the template's lifecycle.
    """
    # These are all available at template definition time and set then
    parse_config: ParseConfig
    render_config: RenderConfig
    # Note: whatever kind of object this is, it needs to be understood by the
    # template loader defined in the template environment.
    # In theory we could make this a typevar, but in practice the overarching
    # ``TemplateIntersectable`` would need to have a typevar within a classvar,
    # which python doesn't currently support.
    resource_locator: object
    # Used primarily for libraries shipping redistributable templates
    # TODO: this should be moved into an EnvConfig object here instead of
    # extracted from it.
    explicit_loader: AsyncTemplateLoader | SyncTemplateLoader | None

    # These are all set during the template loading process, in stages, as
    # increasingly more information is available.
    fieldset: NormalizedFieldset = field(init=False, repr=False)
    total_inclusions: frozenset[TemplateClass] = field(
        init=False, repr=False)
    slot_tree: SlotTreeNode = field(init=False, repr=False)
    # NOTE: a value of None here means that the template HAS NOTHING TO
    # RENDER PREP -- and **not** that the tree hasn't been created yet!
    prerender_tree: PrerenderTreeNode | None = field(init=False, repr=False)

    def expanded_repr(self):
        """This wraps the default repr, including any other non-init
        vars. We chose this as the least-bad way to get reprs to work
        while still having a way to debug the signature manually.
        """
        bare_repr = repr(self)
        noninit_fields: list[str] = []
        for dc_field in fields(self):
            if not dc_field.init:
                noninit_fields.append(dc_field.name)

        to_join = [bare_repr[:-1]]
        for noninit_fieldname in noninit_fields:
            try:
                value = getattr(self, noninit_fieldname)
            except AttributeError:
                value = '<unset>'

            to_join.append(f'{noninit_fieldname}={value}')

        joined = ', '.join(to_join)
        return f'{joined})'



from __future__ import annotations

import functools
import inspect
import logging
import sys
import typing
from collections.abc import Callable
from collections.abc import Collection
from collections.abc import Iterable
from collections.abc import Mapping
from collections.abc import Sequence
from contextlib import contextmanager
from contextvars import ContextVar
from dataclasses import _MISSING_TYPE
from dataclasses import Field
from dataclasses import dataclass
from dataclasses import field
from types import FrameType
from typing import Annotated
from typing import Any
from typing import Literal
from typing import Protocol
from typing import TypedDict
from typing import Unpack
from typing import dataclass_transform
from typing import overload

from docnote import DocnoteConfig
from docnote import Note
from docnote import docnote

from templatey._signature import TemplateSignature
from templatey._types import TemplateClass
from templatey.interpolators import NamedInterpolator
from templatey.modifiers import SegmentModifier
from templatey.parser import InterpolationConfig

logger = logging.getLogger(__name__)

if typing.TYPE_CHECKING:
    from templatey.environments import AsyncTemplateLoader
    from templatey.environments import SyncTemplateLoader

_CLOSURE_ANCHORS: ContextVar[dict[TemplateClass, FrameType]] = ContextVar(
    '_CLOSURE_ANCHORS')


class VariableEscaper(Protocol):

    def __call__(self, value: str) -> str:
        """Variable escaper functions accept a single positional
        argument: the value of the variable to escape. It then does any
        required escaping and returns the final string.
        """
        ...


class ContentVerifier(Protocol):

    def __call__(self, value: str) -> Literal[True]:
        """Content verifier functions accept a single positional
        argument: the value of the content to verify. It does any
        verification, and then returns True if the content was okay,
        or raises BlockedContentValue if the content was not acceptable.

        Note that we raise instead of trying to escape for two reasons:
        1.. We don't really know what to replace it with. This is also
            true with variables, but:
        2.. We expect that content is coming from -- if not trusted,
            then at least authoritative -- sources, and therefore, we
            should fail loudly, because it gives the author a chance to
            correct the problem before it becomes user-facing.
        """
        ...


class InterpolationTransformer[T](Protocol):

    def __call__(
            self,
            value: Annotated[
                T,
                Note(
                    '''The value of the variable or content. A value of
                    ``None`` indicates that the value is intended to be
                    omitted, but you may still provide a fallback
                    instead.
                    ''')]
            ) -> str | None:
        """Interpolation transformers give you a chance to modify the
        rendered result of a particular content or variable value, omit
        it entirely, or provide a fallback for missing values.

        transformers are applied before formatting, escaping, and
        verification, and the result of the transformer is used to
        determine whether or not the value should be included in the
        result. If your transformer returns ``None``, the parameter will
        be completely omitted (including any prefix or suffix).
        """
        ...


@dataclass(slots=True)
class FieldConfig[T]:
    transformer: InterpolationTransformer[T] | None = None


class _DataclassFieldKwargs(TypedDict, total=False):
    init: bool
    repr: bool
    hash: bool | None
    compare: bool
    metadata: Mapping[Any, Any] | None
    kw_only: bool | Literal[_MISSING_TYPE.MISSING]


class _DataclassFieldKwargs14(_DataclassFieldKwargs, TypedDict, total=False):
    doc: str | None


# The following is adapted directly from typeshed. We did some formatting
# updates, and inserted our transformer param.
if sys.version_info >= (3, 14):
    @overload
    def template_field[_T](
            field_config: FieldConfig[_T] = ...,
            /, *,
            default: _T,
            default_factory: Literal[_MISSING_TYPE.MISSING] = ...,
            **field_kwargs: Unpack[_DataclassFieldKwargs14]
            ) -> _T: ...
    @overload
    def template_field[_T](
            field_config: FieldConfig[_T] = ...,
            /, *,
            default: Literal[_MISSING_TYPE.MISSING] = ...,
            default_factory: Callable[[], _T],
            **field_kwargs: Unpack[_DataclassFieldKwargs14]
            ) -> _T: ...
    @overload
    def template_field[_T](
            field_config: FieldConfig[_T] = ...,
            /, *,
            default: Literal[_MISSING_TYPE.MISSING] = ...,
            default_factory: Literal[_MISSING_TYPE.MISSING] = ...,
            **field_kwargs: Unpack[_DataclassFieldKwargs14]
            ) -> Any: ...

# This is technically only valid for >=3.10, but we require that anyways
else:
    @overload
    def template_field[_T](
            field_config: FieldConfig[_T] = ...,
            /, *,
            default: _T,
            default_factory: Literal[_MISSING_TYPE.MISSING] = ...,
            **field_kwargs: Unpack[_DataclassFieldKwargs]
            ) -> _T: ...
    @overload
    def template_field[_T](
            field_config: FieldConfig[_T] = ...,
            /, *,
            default: Literal[_MISSING_TYPE.MISSING] = ...,
            default_factory: Callable[[], _T],
            **field_kwargs: Unpack[_DataclassFieldKwargs]
            ) -> _T: ...
    @overload
    def template_field[_T](
            field_config: FieldConfig[_T] = ...,
            /, *,
            default: Literal[_MISSING_TYPE.MISSING] = ...,
            default_factory: Literal[_MISSING_TYPE.MISSING] = ...,
            **field_kwargs: Unpack[_DataclassFieldKwargs]
            ) -> Any: ...


def template_field(
        field_config: FieldConfig | None = None,
        /, *,
        metadata: Mapping[Any, Any] | None = None,
        **field_kwargs):
    if field_config is None:
        field_config = FieldConfig()

    if metadata is None:
        metadata = {'templatey.field_config': field_config}

    else:
        metadata = {
            **metadata,
            'templatey.field_config': field_config}

    return field(**field_kwargs, metadata=metadata)


# The following is adapted directly from typeshed. We did some formatting
# updates, and inserted our transformer param.
if sys.version_info >= (3, 14):
    @overload
    def param[_T](
            *,
            default: _T,
            default_factory: Literal[_MISSING_TYPE.MISSING] = ...,
            # Deprecated name, but param itself is deprecated
            prerenderer: InterpolationTransformer[_T] | None = None,
            **field_kwargs: Unpack[_DataclassFieldKwargs14],
            ) -> _T: ...
    @overload
    def param[_T](
            *,
            default: Literal[_MISSING_TYPE.MISSING] = ...,
            default_factory: Callable[[], _T],
            # Deprecated name, but param itself is deprecated
            prerenderer: InterpolationTransformer[_T] | None = None,
            **field_kwargs: Unpack[_DataclassFieldKwargs14],
            ) -> _T: ...
    @overload
    def param[_T](
            *,
            default: Literal[_MISSING_TYPE.MISSING] = ...,
            default_factory: Literal[_MISSING_TYPE.MISSING] = ...,
            # Deprecated name, but param itself is deprecated
            prerenderer: InterpolationTransformer[_T] | None = None,
            **field_kwargs: Unpack[_DataclassFieldKwargs14],
            ) -> Any: ...

# This is technically only valid for >=3.10, but we require that anyways
else:
    @overload
    def param[_T](
            *,
            default: _T,
            default_factory: Literal[_MISSING_TYPE.MISSING] = ...,
            # Deprecated name, but param itself is deprecated
            prerenderer: InterpolationTransformer[_T] | None = None,
            **field_kwargs: Unpack[_DataclassFieldKwargs]
            ) -> _T: ...
    @overload
    def param[_T](
            *,
            default: Literal[_MISSING_TYPE.MISSING] = ...,
            default_factory: Callable[[], _T],
            # Deprecated name, but param itself is deprecated
            prerenderer: InterpolationTransformer[_T] | None = None,
            **field_kwargs: Unpack[_DataclassFieldKwargs]
            ) -> _T: ...
    @overload
    def param[_T](
            *,
            default: Literal[_MISSING_TYPE.MISSING] = ...,
            default_factory: Literal[_MISSING_TYPE.MISSING] = ...,
            # Deprecated name, but param itself is deprecated
            prerenderer: InterpolationTransformer[_T] | None = None,
            **field_kwargs: Unpack[_DataclassFieldKwargs]
            ) -> Any: ...


# DEPRECATED. Use ``template_field`` instead.
@docnote(DocnoteConfig(include_in_docs=False))
def param(
        *,
        # Deprecated name, but param itself is deprecated
        prerenderer: InterpolationTransformer | None = None,
        metadata: Mapping[Any, Any] | None = None,
        **field_kwargs):
    field_config = FieldConfig(transformer=prerenderer)

    if metadata is None:
        metadata = {'templatey.field_config': field_config}

    else:
        metadata = {
            **metadata,
            'templatey.field_config': field_config}

    return field(**field_kwargs, metadata=metadata)


class _DataclassKwargs(TypedDict, total=False):
    init: bool
    repr: bool
    eq: bool
    order: bool
    unsafe_hash: bool
    frozen: bool
    match_args: bool
    kw_only: bool
    slots: bool
    weakref_slot: bool


@overload
@docnote(DocnoteConfig(include_in_docs=False))
@dataclass_transform(field_specifiers=(template_field, param, field, Field))
def template[T: type](
        config: TemplateConfig,
        template_resource_locator: object,
        /, *,
        loader: Annotated[
                AsyncTemplateLoader | SyncTemplateLoader | None,
                Note('''Explicit template loaders can be passed to a template
                    instance to bypass the loader declared in the template
                    environment. This is primarily intended as a mechanism for
                    library developers to define redistributable templates as
                    part of their codebases, while not depending on the
                    template loader of the end user's codebase.''')
            ] = None,
        segment_modifiers: Annotated[
                Sequence[SegmentModifier] | None,
                Note('''An ordered sequence of ``SegmentModifier`` instances
                    to apply to every literal string segment of the loaded
                    template text.

                    The modifiers will be applied in the order that they are
                    declared. The first modifier to match the segment will
                    short-circuit the remaining modifiers, regardless of
                    whether or not it applies any changes.''')
            ] = None,
        **dataclass_kwargs: Unpack[_DataclassKwargs]
        ) -> Callable[[T], T]:
    """Deprecated signature using deprecated jack-of-all-trades
    ``TemplateConfig`` instances.
    """
    ...


@overload
@docnote(DocnoteConfig(include_in_docs=False))
@dataclass_transform(field_specifiers=(template_field, param, field, Field))
def template[T: type](
        template_resource_locator: object,
        render_config: RenderConfig,
        /,
        *optional_configs: ParseConfig | EnvConfig,
        **dataclass_kwargs: Unpack[_DataclassKwargs]
        ) -> Callable[[T], T]:
    """Constructs a template class."""
    ...


@dataclass_transform(field_specifiers=(template_field, param, field, Field))
def template[T: type](
        template_config_or_locator: TemplateConfig | object,
        render_config_or_locator: RenderConfig | object,
        /,
        *optional_configs: ParseConfig | EnvConfig,
        loader: AsyncTemplateLoader | SyncTemplateLoader | None = None,
        segment_modifiers: Sequence[SegmentModifier] | None = None,
        **dataclass_kwargs: Unpack[_DataclassKwargs]
        ) -> Callable[[T], T]:
    """This both transforms the decorated class into a stdlib dataclass
    and declares it as a templatey template.

    **Note that unlike the stdlib dataclass decorator, this defaults to
    ``slots=True``.** If you find yourself having problems with
    metaclasses and/or subclassing, you can disable this by passing
    ``slots=False``. Generally speaking, though, this provides a
    free performance benefit. **If weakref support is required, be sure
    to pass ``weakref_slot=True``.
    """
    if isinstance(template_config_or_locator, TemplateConfig):
        legacy_config = template_config_or_locator
        if segment_modifiers is not None:
            segment_modifiers = tuple(segment_modifiers)

        template_resource_locator = render_config_or_locator
        render_config = RenderConfig(
            variable_escaper=legacy_config.variable_escaper,
            content_verifier=legacy_config.content_verifier)
        parse_config = ParseConfig(
            interpolator=legacy_config.interpolator,
            segment_modifiers=segment_modifiers or ())
        env_config = EnvConfig(loader=loader)

    else:
        template_resource_locator = template_config_or_locator

        if not isinstance(render_config_or_locator, RenderConfig):
            raise TypeError('Second arg must be render config!')

        render_config = render_config_or_locator
        parse_config: ParseConfig | None = None
        env_config: EnvConfig | None = None

        for optional_config in optional_configs:
            if isinstance(optional_config, ParseConfig):
                if parse_config is not None:
                    raise TypeError('Cannot have duplicate parse configs!')
                parse_config = optional_config

            elif isinstance(optional_config, EnvConfig):
                if env_config is not None:
                    raise TypeError('Cannot have duplicate env configs!')
                env_config = optional_config

            else:
                raise TypeError('*args must be parse or env configs!')

    # As per docs, we default to using slots, since it makes everything
    # faster
    if 'slots' not in dataclass_kwargs:
        dataclass_kwargs['slots'] = True

    return functools.partial(
        make_template_definition,
        dataclass_kwargs=dataclass_kwargs,
        template_resource_locator=template_resource_locator,
        render_config=render_config,
        parse_config=parse_config,
        env_config=env_config)


# DEPRECATED. Use ``ParseConfig``, ``RenderConfig``, and ``EnvConfig``
# instead.
@docnote(DocnoteConfig(include_in_docs=False))
@dataclass(frozen=True)
class TemplateConfig:
    """Template configs specify how the template and its interpolated
    content should be processed.
    """
    interpolator: Annotated[
        NamedInterpolator,
        Note('''The interpolator determines what characters are used for
            performing interpolations within the template. They can be
            escaped by repeating them, for example ``{{}}`` would be
            a literal ``{}`` with a curly braces interpolator.''')]
    variable_escaper: Annotated[
        VariableEscaper,
        Note('''Variables are always escaped. The variable escaper is
            the callable responsible for performing that escaping. If you
            don't need escaping, there are noop escapers within the prebaked
            template configs that you can use for convenience.''')]
    content_verifier: Annotated[
        ContentVerifier,
        Note('''Content isn't escaped, but it ^^is^^ verified. Content
            verification is a simple process that either succeeds or fails;
            it allows, for example, to allowlist certain HTML tags.''')]


@dataclass(frozen=True, kw_only=True)
class ParseConfig:
    """Parse configs specify how the loaded template text should be
    parsed into a template resource. These can be used to modify all
    segments of a template, change which interpolators to use, etc.
    """
    interpolator: Annotated[
            NamedInterpolator,
            Note('''The interpolator determines what characters are used for
                performing interpolations within the template. They can be
                escaped by repeating them, for example ``{{}}`` would be
                a literal ``{}`` with a curly braces interpolator.''')
        ] = NamedInterpolator.CURLY_BRACES
    segment_modifiers: Annotated[
            tuple[SegmentModifier, ...],
            Note('''An ordered sequence of ``SegmentModifier`` instances
                to apply to every literal string segment of the loaded
                template text.

                The modifiers will be applied in the order that they are
                declared. The first modifier to match the segment will
                short-circuit the remaining modifiers, regardless of
                whether or not it applies any changes.''')
        ] = ()


@dataclass(frozen=True, kw_only=True)
class RenderConfig:
    """Render configs control the behavior of the templatey renderer.
    These can be used to modify variable escaping, content verification,
    etc. These are generally specific to the output format of the
    template (for example, this might be html-specific).
    """
    variable_escaper: Annotated[
        VariableEscaper,
        Note('''Variables are always escaped. The variable escaper is
            the callable responsible for performing that escaping. If you
            don't need escaping, there are noop escapers within the prebaked
            template configs that you can use for convenience.''')]
    content_verifier: Annotated[
        ContentVerifier,
        Note('''Content isn't escaped, but it ^^is^^ verified. Content
            verification is a simple process that either succeeds or fails;
            it allows, for example, to allowlist certain HTML tags.''')]


@dataclass(frozen=True, kw_only=True)
class EnvConfig:
    """Environment configs control the behavior of the render
    environment. They can be used, for example, to specify and explicit
    loader for a template, allowing templates to be packaged within a
    library.
    """
    loader: Annotated[
            AsyncTemplateLoader | SyncTemplateLoader | None,
            Note('''Explicit template loaders can be passed to a template
                instance to bypass the loader declared in the template
                environment. This is primarily intended as a mechanism for
                library developers to define redistributable templates as
                part of their codebases, while not depending on the
                template loader of the end user's codebase.''')
        ] = None


@dataclass_transform(field_specifiers=(template_field, param, field, Field))
def make_template_definition[T: type](
        cls: T,
        *,
        dataclass_kwargs: _DataclassKwargs,
        # Note: needs to be understandable by template loader
        template_resource_locator: object,
        render_config: RenderConfig,
        parse_config: ParseConfig | None,
        env_config: EnvConfig | None,
        ) -> T:
    """Programmatically creates a template definition. Converts the
    requested class into a dataclass, passing along ``dataclass_kwargs``
    to the dataclass constructor. Then performs some templatey-specific
    bookkeeping. Returns the resulting dataclass.
    """
    if env_config is None:
        explicit_loader = None
    else:
        explicit_loader = env_config.loader

    cls = dataclass(**dataclass_kwargs)(cls)
    cls._templatey_signature = TemplateSignature(
        parse_config=parse_config or ParseConfig(),
        render_config=render_config,
        resource_locator=template_resource_locator,
        explicit_loader=explicit_loader)

    closure_anchor = _CLOSURE_ANCHORS.get(None)
    if closure_anchor is not None:
        upstack_frame = _get_first_frame_from_other_module()
        if upstack_frame is not None:
            closure_anchor[cls] = upstack_frame

    return cls


@dataclass(frozen=True, slots=True)
class InjectedValue:
    """This is used by environment functions and complex content to
    indicate that a value is being injected into the template. Use it
    instead of a bare string to preserve an existing interpolation
    config, or to indicate whether verification and/or escaping should
    be applied to the value after conversion to a string.

    Note that, if both are defined, the variable escaper will be called
    first, before the content verifier.
    """
    value: object

    config: InterpolationConfig = field(default_factory=InterpolationConfig)
    use_content_verifier: bool = False
    use_variable_escaper: bool = True

    def __post_init__(self):
        if self.config.prefix is not None or self.config.suffix is not None:
            raise ValueError(
                'Injected values cannot have prefixes nor suffixes. If you '
                + 'need similar behavior, simply add the affix(es) to the '
                + 'iterable returned by the complex content or env function.')


class _ComplexContentBase(Protocol):

    def flatten(
            self,
            dependencies: Annotated[
                Mapping[str, object],
                Note(
                    '''The values of the variables declared as dependencies
                    in the constructor are passed to the call to ``flatten``
                    during rendering.
                    ''')],
            config: Annotated[
                InterpolationConfig,
                Note(
                    '''The interpolation configuration of the content
                    interpolation that the complex content is a member
                    of. Note that neither prefix nor suffix can be passed
                    on to an ``InjectedValue``; they must be manually included
                    in the return value if desired.
                    ''')],
            transformers: Annotated[
                Mapping[str, InterpolationTransformer | None],
                Note(
                    '''If a transformer is defined on a dependency variable,
                    it will be included here; otherwise, the value will be
                    set to ``None``.
                    ''')],
            ) -> Iterable[object | InjectedValue]:
        """Implement this for any instance of complex content.

        First, do whatever content modification you need to, based on
        the dependency variables declared in the constructor. Then,
        if needed, merge in the variable itself using an
        ``InjectedValue``, configuring it as appropriate.

        **Note that the parent interpolation config will be ignored by
        all strings returned by flattening individually.** So if, for
        example, you included a prefix in the content interpolation
        within the template itself, and then passed a ``ComplexContent``
        instance to the template instance, the prefix would be ignored
        completely (unless you do something with it in ``flatten``).

        **Also note that you are responsible for calling the dependency
        variable's ``InterpolationTransformer``. directly,** within your
        implementation of ``flatten``. This affords you the option to
        skip it if desired.

        > Example: noun quantity
        __embed__: 'code/python'
            class NaivePluralContent(ComplexContent):

                def flatten(
                        self,
                        dependencies: Mapping[str, object],
                        config: InterpolationConfig,
                        transformers:
                            Mapping[str, InterpolationTransformer | None],
                        ) -> Iterable[str | InjectedValue]:
                    \"""Pluralizes the name of the provided dependency.
                    For example, ``{'widget': 1}`` will be rendered as
                    "1 widget", but ``{'widget': 2}`` will be rendered as
                    "2 widgets".
                    \"""

                    # Assume only 1 dependency
                    name, value = next(iter(dependencies.items()))

                    if 0 <= value <= 1:
                        return (
                            InjectedValue(
                                value,
                                # This assumes no prefix/suffix
                                config=config,
                                use_content_verifier=False,
                                use_variable_escaper=True),
                            ' ',
                            name)

                    else:
                        return (
                            InjectedValue(
                                value,
                                # This assumes no prefix/suffix
                                config=config,
                                use_content_verifier=False,
                                use_variable_escaper=True),
                            ' ',
                            name,
                            's')
        """
        ...


@dataclass(slots=True, kw_only=True)
class ComplexContent(_ComplexContentBase):
    """Sometimes content isn't as simple as a ``string``. For example,
    content might include variable interpolations. Or you might need
    to modify the content slightly based on the variables -- for
    example, to get subject/verb alignment based on a number, gender
    alignment based on a pronoun, or whatever. ComplexContent gives
    you an escape hatch to do this: simply pass a ComplexContent
    instance as a value instead of a string.
    """

    dependencies: Annotated[
        Collection[str],
        Note(
            '''Complex content dependencies are the **variable** names
            that a piece of complex content depends on. These will be
            passed to the implemented ``flatten`` function during
            rendering.
            ''')]


@contextmanager
def anchor_closure_scope():
    """We strongly recommend against defining templates within a
    closure, as it can cause a number of fragility issues, and just
    generally makes less sense than defining templates at the module
    level. However, if you absolutely must create a new template within
    a closure, you must use ``anchor_closure_scope`` to give the
    templates a known closure scope. Can be used either as a decorator
    or a context:

    > Decorator usage
    __embed__: 'code/python'
        @anchor_closure_scope()
        def my_func():
            # template definition goes here
            ...

    > Context manager usage
    __embed__: 'code/python'
        def my_other_func():
            with anchor_closure_scope():
                # template definition goes here
                ...
    """
    ctx_token = _CLOSURE_ANCHORS.set({})
    try:
        yield
    finally:
        _CLOSURE_ANCHORS.reset(ctx_token)


def _get_first_frame_from_other_module() -> FrameType | None:
    """When templates are created from inside a closure (ex, during
    testing, where this is extremely common), we need access to the
    locals from the closure to resolve type hints. This method relies
    upon ``inspect`` to extract them.

    Note that this can be very sensitive to where, exactly, you put it
    within the templatey code. Always put it as close as possible to
    the public API method, so that the first frame from another module
    coincides with the call to decorate a template class.

    **Note that this is pretty fragile.** Or, put a different way: it
    does exactly what the function name suggest it does: it finds the
    FIRST frame from another module. That doesn't mean we won't return
    to this module; it doesn't mean it's from the actual client library,
    etc. It just means it's the first frame that isn't from this
    module.
    """
    upstack_frame = inspect.currentframe()
    if upstack_frame is None:
        return None
    else:
        # Technically not 100% correct since we might have reloads, but...
        # I mean at that point, good fucking luck with closures.
        this_module = upstack_module = sys.modules[__name__]
        while upstack_module is this_module:
            if upstack_frame is None:
                return None

            upstack_frame = upstack_frame.f_back
            upstack_module = inspect.getmodule(upstack_frame)

    return upstack_frame


@docnote(DocnoteConfig(include_in_docs=False))
def get_closure_locals(template_cls: TemplateClass) -> dict[str, Any] | None:
    """This function -- which is not intended to be part of the public
    API for templatey -- checks first for an anchored closure scope,
    and then (if one is found), checks for a registered frame for the
    passed template class. If everything checks out, we extract the
    locals from that frame and return them; otherwise, we return None.
    """
    closure_anchor = _CLOSURE_ANCHORS.get(None)
    if closure_anchor is not None:
        upstack_frame = closure_anchor.get(template_cls)
        if upstack_frame is not None:
            return upstack_frame.f_locals



from docnote import DocnoteConfig
from docnote import MarkupLang

import templatey.prebaked as prebaked  # noqa: PLR0402
from templatey._types import Content
from templatey._types import DynamicClassSlot
from templatey._types import Slot
from templatey._types import Var
from templatey.environments import RenderEnvironment
from templatey.templates import ComplexContent
from templatey.templates import FieldConfig
from templatey.templates import InjectedValue
from templatey.templates import TemplateConfig
from templatey.templates import anchor_closure_scope
from templatey.templates import param  #  type: ignore  # noqa: F401
from templatey.templates import template
from templatey.templates import template_field

__all__ = [
    'ComplexContent',
    'Content',
    'DynamicClassSlot',
    'FieldConfig',
    'InjectedValue',
    'RenderEnvironment',
    'Slot',
    'TemplateConfig',
    'Var',
    'anchor_closure_scope',
    'prebaked',
    'template',
    'template_field',
]


DOCNOTE_CONFIG = DocnoteConfig(markup_lang=MarkupLang.CLEANCOPY)



from __future__ import annotations

import itertools
import operator
from collections.abc import Iterable
from collections.abc import Sequence
from copy import copy
from dataclasses import KW_ONLY
from dataclasses import InitVar
from dataclasses import dataclass
from dataclasses import field
from dataclasses import fields
from typing import Any
from typing import Self
from typing import cast
from weakref import ref

from templatey._error_collector import ErrorCollector
from templatey._provenance import Provenance
from templatey._provenance import ProvenanceNode
from templatey._renderer import FuncExecutionRequest
from templatey._renderer import TemplateInjection
from templatey._renderer import get_precall_cache_key
from templatey._types import DYNAMIC_TEMPLATE_CLASS
from templatey._types import TemplateClass
from templatey._types import TemplateIntersectable
from templatey._types import TemplateParamsInstance
from templatey._types import create_templatey_id
from templatey.parser import InterpolatedFunctionCall
from templatey.parser import ParsedTemplateResource

type SlotPath = tuple[str, TemplateClass]


class _ProxyDescriptor:
    """This is a bit of a hack. The goal here is to allow dataclasses
    that also inherit from something else (eg list) to include their
    contents in their default-generated repr. The general strategy is
    to use a non-init field as a proxy to the super() repr of the
    object.

    Instead of using a [[descriptor-typed
    field](https://docs.python.org/3/library/dataclasses.html#descriptor-typed-fields)]
    -- which cannot be assigned init=False -- we simply set the field
    as normal, allow the dataclass to be processed, **and then**
    overwrite the field with the descriptor.
    """

    def __get__(self, obj: Any | None, objtype: type | None = None):
        if obj is None:
            return '...'
        elif objtype is None:
            return '...'
        else:
            # Create a disposable shallow copy of the object using the
            # superclass. This makes everything look as expected.
            return objtype.__mro__[1](obj)


# Note: the ordering here is to emphasize the fact that the slot
# name is on the ENCLOSING template, but the slot type is from the
# NESTED template
class PrerenderTreeRoute(tuple[str, TemplateClass, 'PrerenderTreeNode']):
    """An individual route on the prerender tree is defined by the attribute
    name for the slot, the slot type, and the subtree from the slot
    class.

    These are optimized for the non-union case. Traversing the prerender tree
    with union types will result in a bunch of unnecessary comparisons
    against slot names of different slot types.

    Note that prerender tree routes always have a concrete slot name and slot
    type, regardless of whether they're in a pending or concrete tree.
    The reason is simple: in a pending tree, all of the pending classes
    are dead-end nodes, and define their insertion points using just
    the string of the slot name, and nothing else.
    """
    @classmethod
    def new(
            cls,
            slot_name: str,
            slot_type: TemplateClass,
            subtree: PrerenderTreeNode,
            ) -> PrerenderTreeRoute:
        """This seems weird and redundant, but it lets us support kwargs
        for creation, as well as the usual tuple signature.
        """
        return cls((slot_name, slot_type, subtree))

    @property
    def subtree(self) -> PrerenderTreeNode:
        """This is slower than directly accessing the tuple values, but
        it makes for clearer code during tree building, where
        performance isn't quite so critical.
        """
        return self[2]

    @property
    def slot_path(self) -> SlotPath:
        return self[0:2]


@dataclass(slots=True)
class PrerenderTreeNode(list[PrerenderTreeRoute]):
    """The purpose of the prerender tree is to precalculate what sequences of
    getattr() calls we need to traverse to arrive at every instance of a
    particular slot type for a given template, including all nested
    templates.

    **These are optimized for rendering, not for template declaration.**
    Also note that these are optimized for slots that are not declared
    as type unions; type unions will result in a number of unnecessary
    comparisons against the routes of the other slot types in the union.

    The reason this is useful is for batching during rendering. This is
    important for function calls: it allows us to pre-execute all env
    func calls for a template before we start rendering it. In the
    future, it will also serve the same role for discovering the actual
    template types for dynamic slots, allowing us to load the needed
    template types in advance.

    An individual node on the prerender tree is a list of all possible
    attribute names (as ``PrerenderTreeRoute``s) that a particular search
    pass needs to check for a given instance. Note that **all** of the
    attributes must be searched -- hence using an iteration-optimized
    list instead of a mapping.
    """
    routes: InitVar[Iterable[PrerenderTreeRoute] | None] = None

    _: KW_ONLY

    abstract_calls: tuple[InterpolatedFunctionCall, ...]
    dynamic_slot_names: tuple[str, ...]

    id_: int = field(default_factory=create_templatey_id)

    # We use this to make the logic cleaner when navigating trees by paths, but
    # we want the faster performance of the tuple when rendering
    _index_by_slot_path: dict[SlotPath, int] = field(init=False, repr=False)

    # Proxy object for repr; see docnote for _ProxyDescriptor
    _children: list[PrerenderTreeRoute] = field(init=False)

    def __post_init__(
            self,
            routes: Iterable[PrerenderTreeRoute] | None):
        # Explicit instead of super because dataclass on slots breaks super()
        if routes is None:
            list.__init__(self)
        else:
            list.__init__(self, routes)

        self._index_by_slot_path = {
            route.slot_path: index for index, route in enumerate(self)}

    def extract(
            self,
            from_instance: TemplateParamsInstance,
            from_injection: Provenance | None,
            into_injection_backlog: list[TemplateInjection],
            into_precall_backlog: list[FuncExecutionRequest],
            template_preload: dict[TemplateClass, ParsedTemplateResource],
            error_collector: ErrorCollector,
            ) -> None:
        """Extracts all dynamic template injections and function
        executions from the root instance, using ourselves as the tree.
        """
        stack: list[_ExtractionFrame] = [
            _ExtractionFrame(
                active_instance=from_instance,
                active_subtree=self,
                target_subtree_index=0,
                target_instance_index=0,
                wip_provenance=Provenance(
                    (
                        ProvenanceNode(
                            encloser_slot_key='',
                            encloser_slot_index=-1,
                            instance_id=id(from_instance),
                            instance=from_instance),),
                    from_injection=from_injection))]

        while stack:
            frame = stack[-1]
            frame_subtree = frame.active_subtree

            # If the frame is exhausted, we still need to process any dyanamic
            # slots or env func calls on the current instance.
            if frame.exhausted:
                stack.pop()
                frame_instance = frame.active_instance
                frame_provenance = frame.wip_provenance

                for abstract_call in frame_subtree.abstract_calls:
                    args, kwargs = frame_provenance.bind_call_signature(
                        abstract_call,
                        template_preload,
                        error_collector)
                    into_precall_backlog.append(
                        FuncExecutionRequest(
                            abstract_call.name,
                            args=args,
                            kwargs=kwargs,
                            result_key=get_precall_cache_key(
                                frame_provenance, abstract_call),
                            provenance=frame_provenance))

                for dynamic_slot_name in frame_subtree.dynamic_slot_names:
                    into_injection_backlog.extend(
                        (
                            Provenance((
                                ProvenanceNode(
                                    encloser_slot_key='',
                                    encloser_slot_index=-1,
                                    instance_id=id(dynamic_instance),
                                    instance=dynamic_instance),),
                                from_injection=frame_provenance),
                            dynamic_instance)

                        for dynamic_instance
                        in getattr(frame_instance, dynamic_slot_name))

                continue

            slot_route = frame_subtree[frame.target_subtree_index]
            slot_name, slot_type, slot_subtree = slot_route
            target_instance_index = frame.target_instance_index

            # We use the zero-index iteration of the loop
            # to memoize some values on the stack frame.
            # This is, in a way, a nested stack, but we're maintaining
            # the stack state within the _ExtractionFrame.
            if target_instance_index == 0:
                target_instances = getattr(frame.active_instance, slot_name)
                target_instances_count = len(target_instances)

                # Check in advance if there are no target instances at all,
                # and if so, skip the whole thing. This isn't just for
                # performance; the processing logic depends on it.
                if target_instances_count > 0:
                    frame.target_instances_count = target_instances_count
                    frame.target_instances = target_instances
                else:
                    # Note: this is critical! Otherwise we'll infinitely loop.
                    frame.target_subtree_index += 1
                    continue

            else:
                target_instances_count = frame.target_instances_count
                # We've exhausted the target instances; reset the state for
                # the next prerender tree route and then continue.
                if frame.target_instance_index >= target_instances_count:
                    # Note: we're deliberately skipping the target instances
                    # themselves, because it'll just get overwritten the next
                    # time around, so we can save ourselves an operation.
                    frame.target_instances_count = 0
                    frame.target_instance_index = 0
                    # Note: this is critical! Otherwise we'll infinitely loop.
                    frame.target_subtree_index += 1
                    continue

                # We still have some instances to target; normalize the state
                # so that we can operate on them.
                target_instances = frame.target_instances

            instance_to_check = target_instances[target_instance_index]
            frame.target_instance_index += 1
            # Okay, status check: we have our stack frame state configured
            # correctly for the next iteration, and we have target instances
            # to check.
            # We still need to verify that the instances we find actually match
            # the slot type (in case of unions), but if we find a match, we'll
            # need to add a new frame to the stack.
            # Note: exact match here; not subclassing! Subclassing breaks too
            # many things, so we don't support it.
            if type(instance_to_check) is slot_type:
                stack.append(_ExtractionFrame(
                    active_instance=instance_to_check,
                    active_subtree=slot_subtree,
                    target_subtree_index=0,
                    target_instance_index=0,
                    wip_provenance=Provenance(
                        (
                            *frame.wip_provenance.slotpath,
                            ProvenanceNode(
                                encloser_slot_key=slot_name,
                                encloser_slot_index=target_instance_index,
                                instance_id=id(instance_to_check),
                                instance=instance_to_check)),
                        from_injection=from_injection)))

    def empty_clone(
            self,
            *,
            fields_to_skip: set[str] | frozenset[str] = frozenset()
            ) -> Self:
        """This creates a clone of the node without any routes. Useful
        for merging and copying, where you need to do some manual
        transform of the content.

        Note that this is almost the same as dataclasses.replace, with
        the exception that we create shallow copies of attributes
        instead of preserving them.
        """
        kwargs = {}
        for dc_field in fields(self):
            if dc_field.init and dc_field.name not in fields_to_skip:
                # Note: the copy here is important for any mutable values,
                # notably the insertion_slot_names.
                kwargs[dc_field.name] = copy(getattr(self, dc_field.name))

        return type(self)(**kwargs)

    def merge_fields_only(
            self,
            other: PrerenderTreeNode,
            *,
            fields_to_skip: set[str] | frozenset[str] = frozenset()):
        """Updates the current node, merging in all non-init field
        values from other, using |=. Only merges values that exist on
        the current node, allowing for transformation between pending
        and concrete node types.

        Leaves the ID of the current node unchanged.
        """
        # Always skip ``id_`` -- always preserve the original ID!
        fields_to_skip = fields_to_skip | {'id_'}
        missing = object()

        for dc_field in fields(self):
            if dc_field.init and dc_field.name not in fields_to_skip:
                current_value = getattr(self, dc_field.name)
                other_value = getattr(other, dc_field.name, missing)

                if other_value is not missing:
                    setattr(
                        self,
                        dc_field.name,
                        operator.ior(current_value, other_value))

    def append(self, route: PrerenderTreeRoute):
        """This has slightly different semantics to normal list
        appending if you're trying to append a duplicate slot path:
        ++  if it's identical to the existing path (ie, it targets the
            same node), we do nothing
        ++  if it's not identical, we error
        """
        slot_path = (route[0], route[1])
        slot_index = self._index_by_slot_path.get(slot_path)

        if slot_index is not None:
            if self[slot_index].subtree is route.subtree:
                return
            else:
                raise ValueError(
                    'Templatey internal error: attempt to append duplicate '
                    + 'slot name for same slot type, but a different subtree! '
                    + 'Please search for / report issue to github along with '
                    + 'a traceback.')

        list.append(self, route)
        self._index_by_slot_path[slot_path] = len(self) - 1

    def has_route_for(
            self,
            slot_name: str,
            slot_type: TemplateClass
            ) -> bool:
        return (slot_name, slot_type) in self._index_by_slot_path

    def get_route_for(
            self,
            slot_name: str,
            slot_type: TemplateClass
            ) -> PrerenderTreeRoute:
        return self[self._index_by_slot_path[(slot_name, slot_type)]]

    def __truediv__(self, other: SlotPath) -> PrerenderTreeNode:
        """A utility method for tree traversal. Only intended for use
        in debugging and testing; not optimized for production use.
        """
        return self.get_route_for(*other)[2]

    def rewrite_route_for(
            self,
            slot_name: str,
            slot_type: TemplateClass,
            new_route: PrerenderTreeRoute
            ) -> None:
        dest_index = self._index_by_slot_path[(slot_name, slot_type)]
        list.__setitem__(self, dest_index, new_route)

    def stringify(
            self,
            *,
            depth=0,
            _encountered_ids: frozenset[int] = frozenset()
            ) -> str:
        """Creates a pretty-print-style string representation of the
        node and all its nested nodes, recursively.
        """
        indentation = '    ' * depth

        to_join = []
        for dc_field in fields(self):
            if dc_field.init:
                field_name = dc_field.name
                to_join.append(
                    f'{indentation}{field_name}: {getattr(self, field_name)}')

        if self.id_ in _encountered_ids:
            to_join.append(
                f'{indentation}... Recursion detected; omitting subtrees.')

        else:
            for route in self:
                to_join.append(
                    f'{indentation}++  {route[0:2]}')
                to_join.append(route[2].stringify(
                    depth=depth + 1,
                    _encountered_ids=_encountered_ids | {self.id_}))

        return '\n'.join(to_join)

    def is_equivalent(
            self,
            other: PrerenderTreeNode,
            *,
            _previous_encounters: dict[int, int] | None = None
            ) -> bool:
        """This compares two prerender tree nodes recursively, ignoring IDs.
        If they are otherwise identical -- including the structure of
        any recursive loops -- returns True.

        Note that this is optimized for maintainability and not
        performance. Its primary intended use is in testing.
        """
        previous_encounters: dict[int, int]
        if _previous_encounters is None:
            previous_encounters = {}
        else:
            # Note that the copy here is important because otherwise we'd be
            # mutating state even in other tree branches which might not have
            # encountered us (since we reuse this in the recursive case)
            previous_encounters = {**_previous_encounters}

        # These two cases are trivial: first, if the types aren't the same,
        # they're not equivalent, period. Second, if we already encountered
        # that node's ID, that means we're in a recursive loop, and the two
        # IDs must match.
        if type(self) is not type(other):
            return False
        if other.id_ in previous_encounters:
            return previous_encounters[other.id_] == self.id_

        # First check all the fields, since this should in theory be quick.
        # Yes we could do a bool for this, but doing it as a dict makes it
        # easier to add in temporary print debugs if required
        fields_match: dict[str, bool] = {}
        for dc_field in fields(self):
            field_name = dc_field.name
            if dc_field.init and field_name != 'id_':
                fields_match[field_name] = (
                    getattr(self, field_name) == getattr(other, field_name))
        if not all(fields_match.values()):
            return False

        # Now make sure the routes are the same. This lets us simplify the
        # recursive comparison logic; we don't need any checks to make sure
        # that there weren't any leftover routes on either self or other.
        # It also lets us short-circuit without recursion if they don't match,
        # but that's not the primary motivation behind it.
        self_routes = set(self._index_by_slot_path)
        other_routes = set(other._index_by_slot_path)
        if self_routes != other_routes:
            return False

        # Now finally we're getting to recursive territory.
        previous_encounters[other.id_] = self.id_
        for slot_path in self._index_by_slot_path:
            self_index = self._index_by_slot_path[slot_path]
            self_subtree = self[self_index].subtree
            other_index = other._index_by_slot_path[slot_path]
            other_subtree = other[other_index].subtree

            # Recursion is glorious as long as you don't need to worry about
            # performance! Too bad we can't use it during rendering (because
            # it's way too slow)
            if not self_subtree.is_equivalent(
                other_subtree,
                _previous_encounters=previous_encounters
            ):
                return False

        return True

    def extend(self, routes: Iterable[PrerenderTreeRoute]):
        raise ZeroDivisionError('Templatey internal error: not implemented!')

    def __delitem__(self, *args, **kwargs):
        raise ZeroDivisionError('Templatey internal error: not implemented!')

    def __setitem__(self, *args, **kwargs):
        raise ZeroDivisionError('Templatey internal error: not implemented!')

    def clear(self, *args, **kwargs):
        raise ZeroDivisionError('Templatey internal error: not implemented!')

    def copy(self, *args, **kwargs):
        raise ZeroDivisionError('Templatey internal error: not implemented!')

    def insert(self, *args, **kwargs):
        raise ZeroDivisionError('Templatey internal error: not implemented!')

    def pop(self, *args, **kwargs):
        raise ZeroDivisionError('Templatey internal error: not implemented!')

    def remove(self, *args, **kwargs):
        raise ZeroDivisionError('Templatey internal error: not implemented!')

    def reverse(self, *args, **kwargs):
        raise ZeroDivisionError('Templatey internal error: not implemented!')

    def __imul__(self, *args, **kwargs):
        raise ZeroDivisionError('Templatey internal error: not implemented!')

    def __iadd__(self, *args, **kwargs):
        raise ZeroDivisionError('Templatey internal error: not implemented!')
PrerenderTreeNode._children = _ProxyDescriptor()  # type: ignore


@dataclass(slots=True)
class _ExtractionFrame:
    """
    """
    active_instance: TemplateParamsInstance
    active_subtree: PrerenderTreeNode
    target_subtree_index: int
    target_instance_index: int
    target_instances_count: int = field(kw_only=True, default=0)
    target_instances: Sequence[TemplateParamsInstance] = field(
        kw_only=True, init=False)
    wip_provenance: Provenance

    _active_subtree_len: int = field(init=False, repr=False, compare=False)

    def __post_init__(self):
        self._active_subtree_len = len(self.active_subtree)

    @property
    def exhausted(self) -> bool:
        return self.target_subtree_index >= self._active_subtree_len


type _SlotTreeRoute = tuple[str, TemplateClass, SlotTreeNode]


@dataclass(slots=True, weakref_slot=True)
class SlotTreeNode(list[_SlotTreeRoute]):
    """Slot trees paint the full picture of all slots on all
    children of a particular root template class. They are not culled,
    and therefore not optimized for rendering. However, they can be
    generated as soon as template recursion totality is achieved -- ie,
    before actually loading the template bodies (in contrast to the
    prerender tree, which needs to know function invocations, and
    therefore requires template loading).

    These are used as an intermediate step to building a prerender tree.
    The challenge here is that we need to avoid infinite recursion when
    resolving mutually-recursive loops of slot dependencies. So we need
    a way that we can short-circuit tree population any time we
    encounter a slot with an already encountered slot class.

    These frames allow for recursion detection by reducing the problem
    to something (metaphorically) similar to this:
    >
    __embed__: 'code/python'
        def recursively_add_slot_classes(
                parent,
                all_slots: set[TemplateClass]):
            for slot_class in parent:
                if slot_class in all_slots:
                    continue
                else:
                    recursively_add_slot_classes(slot_class, all_slots)

    Although this requires a little bit of duplicated effort when we
    then need to build the trees for other slot classes along the
    recursion path, it massively simplifies the code to build them.

    These are not meant to be created directly; instead, use the two
    helpers, ``make_root`` and ``add_child``.
    """
    # Note: these are (deliberately) redundant with the nodepath that led us
    # here, and included purely for convenience
    slot_name: str
    slot_cls: TemplateClass

    # Note that this does NOT include the current node
    _nodepath_parents: tuple[tuple[*SlotPath, ref[SlotTreeNode]], ...] = field(
        repr=False)

    # These are always self-referential, so we use a weakref to avoid literally
    # always needing complicated GC
    first_encounters: dict[TemplateClass, ref[SlotTreeNode]] = field(
        repr=False)

    dynamic_slot_names: set[str]
    id_: int = field(default_factory=create_templatey_id)

    # Proxy object for repr; see docnote for _ProxyDescriptor
    _children: list[_SlotTreeRoute] = field(init=False)

    def __truediv__(self, other: SlotPath) -> SlotTreeNode:
        """A utility method for tree traversal. Only intended for use
        in debugging and testing; not optimized for production use.
        """
        for slot_name, slot_cls, slot_node in self:
            if (slot_name, slot_cls) == other:
                return slot_node

        raise LookupError('No such slot path!', other)

    def get_all_nodepath_inclusions(
            self,
            offset: SlotTreeNode | None = None
            ) -> Iterable[tuple[*SlotPath, SlotTreeNode]]:
        """Provides a view into all nodepath parents, plus the current
        node, in order. If offset is passed, start immediately after the
        passed node.
        """
        if offset is None:
            yield from self.nodepath_parents
        else:
            offset_encountered = False
            for slot_name, slot_cls, slot_tree_node in self.nodepath_parents:
                if offset_encountered:
                    yield (slot_name, slot_cls, slot_tree_node)

                if slot_tree_node is offset:
                    offset_encountered = True

            if not offset_encountered:
                raise ValueError(
                    'Nodepath inclusion offset not found within parents!',
                    self, offset)

        yield (self.slot_name, self.slot_cls, self)

    @property
    def nodepath_parents(
            self
            ) -> tuple[tuple[*SlotPath, SlotTreeNode], ...]:
        """This dereferences the nodepath parents.
        """
        # Note: the cast() here is because pyright doesn't pick up on the
        # any() we do in just a second to fix it, and the explicit typecast
        # is because otherwise pyright assigns the type too broadly and it
        # no longer matches the return signature
        retval: tuple[tuple[*SlotPath, SlotTreeNode], ...] = tuple(
            (slot_name, slot_cls, cast(SlotTreeNode, noderef()))
            for slot_name, slot_cls, noderef in self._nodepath_parents)
        if any(segment[2] is None for segment in retval):
            raise RuntimeError(
                'Impossible branch: prematurely GCd nodepath parents!',
                self)

        return retval

    def check_recursion(
            self,
            template_cls: TemplateClass
            ) -> SlotTreeNode | None:
        """Checks for the passed ``template_cls`` in the history. If
        found, returns a it (after dereferencing). If not found, returns
        None. If the reference is no longer valid, raises.
        """
        if template_cls in self.first_encounters:
            node = self.first_encounters[template_cls]()
            if node is None:
                raise RuntimeError(
                    'Impossible branch: prematurely GCd slot tree '
                    + 'node!', self)

            return node

        return None

    def add_child(
            self,
            slot_name: str,
            slot_cls: TemplateClass
            ) -> SlotTreeNode:
        """Creates a child node and appends it to self, then returns it.
        """
        new_child = SlotTreeNode(
            slot_name=slot_name,
            slot_cls=slot_cls,
            first_encounters={**self.first_encounters},
            dynamic_slot_names=set(),
            _nodepath_parents=(
                *self._nodepath_parents, (slot_name, slot_cls, ref(self))),)

        if (
            slot_cls is not DYNAMIC_TEMPLATE_CLASS
            and slot_cls not in new_child.first_encounters
        ):
            new_child.first_encounters[slot_cls] = ref(new_child)

        self.append((slot_name, slot_cls, new_child))
        return new_child

    @classmethod
    def make_root(
            cls,
            template_cls: TemplateClass,
            ) -> SlotTreeNode:
        new_root = cls(
            # Hacky, but... easier than needing to always check for Nones.
            slot_name='',
            slot_cls=template_cls,
            first_encounters={},
            dynamic_slot_names=set(),
            _nodepath_parents=(),)
        new_root.first_encounters[template_cls] = ref(new_root)
        return new_root

    def distill_prerender_tree(
            self,
            template_preload: dict[TemplateClass, ParsedTemplateResource],
            ) -> PrerenderTreeNode | None:
        """Call this on the root of the slot tree to create a prerender
        tree based on the template preload. This will correctly cull all
        non-relevant slot paths (ie, any slot paths with neither dynamic
        slot classes nor env func calls) while keeping recursion loops
        intact.

        Overview of the algorithm:
        ++  iterate over the whole pending tree, from shallowest node to
            deepest node. this allows you to guarantee that you're finding
            the deepest recursions first.
        ++  keep track of the ``postrecursion_descendants`` separately.
            these get updated during the transformation.
            probably do this as an {id: set[]} construct, that way it's
            really easy to update.
        ++  every time you encounter a pending node with recursion loop
            sources:
            ++  for each recursion loop source
            ++  ``get_all_nodepath_inclusions`` (probably will want to add
                an offset parameter) FROM the recursion target UNTIL the
                recursion source (inclusive) -- ie, all intermediate nodes
                within the recursion loop
            ++  add the current pending node's postrecursion descendants
                (plus the slot class for the pending node) to that
                intermediate node's postrecursion descendants
        ++  every time you encounter a node, add the nonrecursive descendants
            to the postrecursion descendants to get the total descendants
            for that node
        ++  if the current slot class matches the target, include, regardless
            of how many children it has.
        ++  for each child:
            ++  if the target slot class isn't included in the child's
                total descendants, cull the child (including all its children)
        ++  if the current slot class does not match the target, AND all
            its children were culled, cull the current pending node in its
            entirety
        """
        inclusions_with_precall: set[TemplateClass] = {
            template_cls
            for template_cls, parsed_template in template_preload.items()
            if parsed_template.function_calls}
        inclusions_with_dynacls: set[TemplateClass] = {
            template_cls
            for template_cls in template_preload
            if cast(
                type[TemplateIntersectable], template_cls
            # Note: this is a little fragile; if you start abusing the preload
            # to do things it isn't meant to do, you might break things (if
            # the other templates haven't been loaded yet)
            )._templatey_signature.fieldset.dynamic_class_slot_names}
        target_slot_classes = inclusions_with_precall | inclusions_with_dynacls

        # We use this to detect and resolve recursion loops
        provisioned_node_by_id: dict[int, PrerenderTreeNode] = {}

        # Setting the root node to None in advance makes it trivial to detect
        # an empty result. Or more accurately, it means we don't have to detect
        # anything; we can just use the value directly
        root_frame = _PrerenderTreeBuilderFrame.provision(
            slot_tree_node=self,
            template_preload=template_preload,
            target_slot_classes=target_slot_classes,
            provisioned_node_by_id=provisioned_node_by_id)
        # Early return: if we don't need a frame for the root, then it means
        # that the entire tree ~~is boring~~ has no dynamic classes and no
        # precalls and therefore no prerender tree
        if root_frame is None:
            return

        root_node = root_frame.pending_prerender_tree_node
        stack: list[_PrerenderTreeBuilderFrame] = [root_frame]
        while stack:
            frame = stack[-1]

            if frame.exhausted:
                stack.pop()
                continue

            slot_name, slot_cls, nested_slot_tree_node = frame.advance()
            recursive_node = provisioned_node_by_id.get(
                nested_slot_tree_node.id_)

            if recursive_node is None:
                deeper_frame = _PrerenderTreeBuilderFrame.provision(
                    slot_tree_node=nested_slot_tree_node,
                    template_preload=template_preload,
                    target_slot_classes=target_slot_classes,
                    provisioned_node_by_id=provisioned_node_by_id)
                if deeper_frame is not None:
                    frame.pending_prerender_tree_node.append(
                        PrerenderTreeRoute.new(
                            slot_name,
                            slot_cls,
                            deeper_frame.pending_prerender_tree_node))
                    stack.append(deeper_frame)

            # This could be either a direct/trivial recursion or a recursion
            # chain; either way it'll be resolved.
            else:
                frame.pending_prerender_tree_node.append(
                    PrerenderTreeRoute.new(
                        slot_name, slot_cls, recursive_node))

        return root_node

    def extend(self, *args, **kwargs):
        raise ZeroDivisionError('Templatey internal error: not implemented!')

    def __delitem__(self, *args, **kwargs):
        raise ZeroDivisionError('Templatey internal error: not implemented!')

    def __setitem__(self, *args, **kwargs):
        raise ZeroDivisionError('Templatey internal error: not implemented!')

    def clear(self, *args, **kwargs):
        raise ZeroDivisionError('Templatey internal error: not implemented!')

    def copy(self, *args, **kwargs):
        raise ZeroDivisionError('Templatey internal error: not implemented!')

    def insert(self, *args, **kwargs):
        raise ZeroDivisionError('Templatey internal error: not implemented!')

    def pop(self, *args, **kwargs):
        raise ZeroDivisionError('Templatey internal error: not implemented!')

    def remove(self, *args, **kwargs):
        raise ZeroDivisionError('Templatey internal error: not implemented!')

    def reverse(self, *args, **kwargs):
        raise ZeroDivisionError('Templatey internal error: not implemented!')

    def __imul__(self, *args, **kwargs):
        raise ZeroDivisionError('Templatey internal error: not implemented!')

    def __iadd__(self, *args, **kwargs):
        raise ZeroDivisionError('Templatey internal error: not implemented!')
SlotTreeNode._children = _ProxyDescriptor()  # type: ignore


@dataclass(slots=True)
class _PrerenderTreeBuilderFrame:
    """``_PrerenderTreeBuilderFrame`` instances are responsible for
    building up a prerender tree from a slot tree.
    """
    slot_tree_node: SlotTreeNode
    pending_prerender_tree_node: PrerenderTreeNode
    child_index: int = 0

    @property
    def exhausted(self) -> bool:
        return self.child_index >= len(self.slot_tree_node)

    def advance(self) -> _SlotTreeRoute:
        next_route = self.slot_tree_node[self.child_index]
        self.child_index += 1
        return next_route

    @classmethod
    def provision(
            cls,
            slot_tree_node: SlotTreeNode,
            template_preload: dict[TemplateClass, ParsedTemplateResource],
            target_slot_classes: set[TemplateClass],
            provisioned_node_by_id: dict[int, PrerenderTreeNode],
            ) -> _PrerenderTreeBuilderFrame | None:
        """Call this to create a new frame if, and only if, one is
        required. Returns the created frame if one is needed; otherwise
        (ie, if the branch should be culled) returns None.
        """
        slot_cls = slot_tree_node.slot_cls
        slot_signature = cast(
            type[TemplateIntersectable], slot_cls)._templatey_signature

        # This is pretty simple. We've already gone to the effort of
        # calculating the total inclusions for everything. Now we can just
        # check to see if there's any overlap with the target classes; if so,
        # we need to keep it, but if not, well, we can discard.
        # This fully encapsulates all of the complicated logic checks we'd
        # otherwise be doing into a simple set intersection.
        if target_slot_classes & slot_signature.total_inclusions:
            parsed_template = template_preload[slot_cls]

            new_node = PrerenderTreeNode(
                # This is partly for convenience (by keeping the
                # IDs the same it's easier to see what's going on)
                # but is also critical for resolving recursion
                # (see below)
                id_=slot_tree_node.id_,
                abstract_calls=tuple(itertools.chain.from_iterable(
                    parsed_template.function_calls.values())),
                dynamic_slot_names=tuple(
                    # Sorting here to maintain consistent ordering; can be
                    # helpful with tests
                    sorted(slot_tree_node.dynamic_slot_names)))
            provisioned_node_by_id[new_node.id_] = new_node

            return cls(
                slot_tree_node=slot_tree_node,
                pending_prerender_tree_node=new_node)

        # Implicit second case: if there's no overlap between the total
        # inclusions and the target classes, then we never update the new_node,
        # and this returns None (relevant for the root node) and doesn't append
        # the route to the parent (relevant for other nodes) -- therefore
        # culling the node and its children.
        return None


@dataclass(slots=True)
class _SlotTreeBuilderFrame:
    """``_SlotTreeBuilderFrame`` instances are used to build up a
    slot tree, which is then used to construct the prerender tree.
    """
    slot_cls: TemplateClass
    pending_slot_tree_node: SlotTreeNode
    remaining_nested_slots: list[SlotPath]
    dynamic_slot_names: frozenset[str]

    @classmethod
    def from_slot_cls(
            cls,
            slot_cls: TemplateClass,
            pending_slot_tree_node: SlotTreeNode,
            ) -> _SlotTreeBuilderFrame:
        """Constructs a new slot tree builder frame for the passed
        template class. **Note that the pending node is for the passed
        slot_cls!**
        """
        fieldset = cast(
            type[TemplateIntersectable], slot_cls
        )._templatey_signature.fieldset
        return cls(
            slot_cls=slot_cls,
            pending_slot_tree_node=pending_slot_tree_node,
            remaining_nested_slots=list(fieldset.slotpaths),
            dynamic_slot_names=fieldset.dynamic_class_slot_names)

    @property
    def exhausted(self) -> bool:
        return not self.remaining_nested_slots


def build_slot_tree(
        template_cls: TemplateClass
        ) -> SlotTreeNode:
    """
    """
    # First we need to build up the pending slot tree, which contains all of
    # the nested template classes with no filtering or culling
    root_node = SlotTreeNode.make_root(template_cls)
    stack: list[_SlotTreeBuilderFrame] = [
        _SlotTreeBuilderFrame.from_slot_cls(template_cls, root_node)]

    while stack:
        frame = stack[-1]
        if frame.exhausted:
            stack.pop()
            # Note: dynamic template classes neither add a frame to the stack,
            # nor are they stored like explicit classes; instead, they simply
            # become string values stored on the pending node.
            # Do this during exhaustion so it only happens once per frame
            # instead of once per iteration.
            frame.pending_slot_tree_node.dynamic_slot_names.update(
                frame.dynamic_slot_names)
            continue

        nested_slot_name, nested_slot_cls = frame.remaining_nested_slots.pop()

        # In the simple recursion case -- a template defines a slot of its
        # own class -- we can immediately create a reference loop without
        # any pomp nor circumstance.
        if nested_slot_cls is frame.slot_cls:
            trivially_recursive_slot_route = (
                nested_slot_name,
                nested_slot_cls,
                frame.pending_slot_tree_node)
            frame.pending_slot_tree_node.append(trivially_recursive_slot_route)
            # Note that we don't need to add this to the recursion sources,
            # because trivial recursion can't influence the culling of the tree

        # In the slightly more complicated recursion case -- a template defines
        # a slot of an already-encountered class -- we just need to retrieve
        # the previous node.
        elif (
            first_encounter
                := frame.pending_slot_tree_node.check_recursion(
                    nested_slot_cls)
        ) is not None:
            mutually_recursive_slot_route = (
                nested_slot_name,
                nested_slot_cls,
                first_encounter)
            frame.pending_slot_tree_node.append(mutually_recursive_slot_route)

        # In the non-recursive case, we need to descend deeper into the
        # dependency graph.
        else:
            next_node = frame.pending_slot_tree_node.add_child(
                nested_slot_name, nested_slot_cls)
            next_frame = _SlotTreeBuilderFrame.from_slot_cls(
                nested_slot_cls, next_node)
            stack.append(next_frame)

    return root_node



import inspect
import typing
from collections.abc import Awaitable
from collections.abc import Callable
from collections.abc import Iterable
from collections.abc import Sequence
from dataclasses import dataclass
from functools import partial
from typing import Annotated
from typing import Literal
from typing import Optional
from typing import Protocol
from typing import cast
from typing import runtime_checkable

try:
    import anyio
    from anyio import create_task_group
except ImportError:
    if typing.TYPE_CHECKING:
        import anyio
        from anyio import create_task_group
from docnote import Note

from templatey._bootstrapping import PARSED_EMPTY_TEMPLATE
from templatey._bootstrapping import EmptyTemplate
from templatey._error_collector import ErrorCollector
from templatey._finalizers import finalize_signature
from templatey._renderer import FuncExecutionRequest
from templatey._renderer import FuncExecutionResult
from templatey._renderer import PrecallCacheKey
from templatey._renderer import RenderContext
from templatey._renderer import TemplateInjection
from templatey._renderer import render_driver
from templatey._signature import TemplateSignature
from templatey._types import TemplateClass
from templatey._types import TemplateIntersectable
from templatey._types import TemplateParamsInstance
from templatey._types import is_template_instance
from templatey.exceptions import MismatchedRenderColor
from templatey.exceptions import MismatchedTemplateEnvironment
from templatey.exceptions import MismatchedTemplateSignature
from templatey.exceptions import TemplateyException
from templatey.parser import ParsedTemplateResource
from templatey.parser import parse
from templatey.templates import InjectedValue

# Note: strings here will be escaped. InjectedValues may decide whether or not
# escaping should be applied. Nested templates will not be escaped.
EnvFunction = Callable[
    ..., Sequence[str | TemplateParamsInstance | InjectedValue]]
EnvFunctionAsync = Callable[
    ..., Awaitable[Sequence[str | TemplateParamsInstance | InjectedValue]]]


@dataclass(frozen=True, slots=True)
class _TemplateFunctionContainer[F: EnvFunction | EnvFunctionAsync]:
    """
    """
    name: str
    function: F
    signature: inspect.Signature
    is_async: bool


@runtime_checkable
class SyncTemplateLoader[L: object](Protocol):

    def load_sync(
            self,
            template: type[TemplateParamsInstance],
            template_resource_locator: L
            ) -> str:
        """This is responsible for loading the actual template text,
        based on the passed resource locator.
        """
        ...


@runtime_checkable
class AsyncTemplateLoader[L: object](Protocol):

    async def load_async(
            self,
            template: type[TemplateParamsInstance],
            template_resource_locator: L
            ) -> str:
        """This is responsible for loading the actual template text,
        based on the passed resource locator.
        """
        ...


class RenderEnvironment:
    _parsed_template_cache: dict[
        type[TemplateParamsInstance], ParsedTemplateResource]
    _env_functions: dict[str, _TemplateFunctionContainer]
    # We use this to prevent registering template functions after any calls
    # to load() have been made, because it would result in different template
    # functions per template
    _has_loaded_any_template: bool
    _template_loader: SyncTemplateLoader | AsyncTemplateLoader
    strict_interpolation_validation: bool

    def __init__(
            self,
            template_loader: SyncTemplateLoader | AsyncTemplateLoader,
            env_functions:
                Optional[Iterable[EnvFunction | EnvFunctionAsync]] = None,
            # If True, this will make sure that the template interface exactly
            # matches the template text. If False, this will just make sure
            # that the template interface is at least sufficient for the
            # template text.
            strict_interpolation_validation: bool = True):
        self.strict_interpolation_validation = strict_interpolation_validation
        self._has_loaded_any_template = False

        self._env_functions = {}
        if env_functions is not None:
            for function in env_functions:
                self.register_env_function(function)

        self._can_load_sync = isinstance(template_loader, SyncTemplateLoader)
        self._can_load_async = isinstance(template_loader, AsyncTemplateLoader)
        self._template_loader = template_loader
        self._parsed_template_cache = {}

    def register_env_function(
            self,
            env_function: EnvFunction | EnvFunctionAsync,
            *,
            force_async: bool = False,
            with_name: str | None = None):
        """Manually register an environment function with the render
        environment, instead of passing it in to the environment
        constructor.

        This can be used to force a function to be registerd as async,
        in case it was not inferred as such, by passing
        ``force_async=True``.

        Normally, registered functions are assigned their __name__ as
        the function name; manual registration can also be used to
        override this behavior via the ``with_name`` parameter.
        """
        if self._has_loaded_any_template:
            raise TemplateyException(
                'To prevent having different template functions per template, '
                + 'you cannot register new template functions after loading '
                + 'any templates in an environment.')

        if with_name is None:
            function_name = env_function.__name__
        else:
            function_name = with_name

        self._env_functions[function_name] = _TemplateFunctionContainer(
            name=function_name,
            function=env_function,
            signature=inspect.signature(env_function),
            is_async=force_async or _infer_asyncness(env_function))

    async def load_async(
            self,
            template: type[TemplateParamsInstance],
            *,
            override_validation_strictness: None | bool = None,
            force_reload: bool = False,
            preload: Annotated[
                    dict[TemplateClass, ParsedTemplateResource] | None,
                    Note('''If desired, pass ``preload`` to recover all of the
                        parsed template resources, including dependencies.

                        If omitted, the dependent resources will still be
                        loaded and cached, but there won't be any guarantees
                        against them subsequently being evicted before the
                        call to load returns.''')
                ] = None
            ) -> ParsedTemplateResource:
        """Loads a template resource from a TemplateParamsInstance
        class. Caches it within the environment and returns the
        resulting ParsedTemplateResource.

        If force_reload is True, bypasses the cache.

        Note that this will also load any and all other templates that
        might be required (via nested slots) to render the passed
        template. These will be cached, but not directly returned.
        """
        # Note: doing this first for thread safety in the sync case, preventing
        # any new template functions from being registered.
        self._has_loaded_any_template = True
        # Note: in this case, we can bypass iterally everything.
        if template is EmptyTemplate:
            await anyio.sleep(0)
            return PARSED_EMPTY_TEMPLATE

        template_xable = cast(type[TemplateIntersectable], template)
        signature = template_xable._templatey_signature
        template_loader = self._get_loader(
            signature,
            self._can_load_async,
            AsyncTemplateLoader)

        # We're doing this sleep just in case everything was cached, and we
        # never await a single loader.
        await anyio.sleep(0)
        with finalize_signature(
            signature,
            template,
            force_reload=force_reload,
            preload=preload,
            parse_cache=self._parsed_template_cache
        ) as sig_finalizer:
            async with create_task_group() as task_group:
                for required_template_cls in sig_finalizer.required_loads:
                    task_group.start_soon(
                        self._wrap_load_async,
                        template_loader,
                        required_template_cls,
                        cast(
                            type[TemplateIntersectable], required_template_cls
                        )._templatey_signature,
                        sig_finalizer.preload,
                        override_validation_strictness)

        return sig_finalizer.target_resource

    async def _wrap_load_async(
            self,
            template_loader: AsyncTemplateLoader,
            template_cls: TemplateClass,
            template_signature: TemplateSignature,
            preload: dict[TemplateClass, ParsedTemplateResource],
            override_validation_strictness: bool | None
            ) -> None:
        """This wraps load_async so that we can parallelize the template
        loading.
        """
        requirement_text = await template_loader.load_async(
            template_cls,
            template_signature.resource_locator)
        parsed_requirement_template = self._parse_and_validate(
            template_cls,
            template_signature,
            requirement_text,
            override_validation_strictness)

        preload[template_cls] = parsed_requirement_template

    def load_sync(
            self,
            template: type[TemplateParamsInstance],
            *,
            override_validation_strictness: None | bool = None,
            force_reload: bool = False,
            preload: Annotated[
                    dict[TemplateClass, ParsedTemplateResource] | None,
                    Note('''If desired, pass ``preload`` to recover all of the
                        parsed template resources, including dependencies.

                        If omitted, the dependent resources will still be
                        loaded and cached, but there won't be any guarantees
                        against them subsequently being evicted before the
                        call to load returns.''')
                ] = None
            ) -> ParsedTemplateResource:
        """Loads a template resource from a TemplateParamsInstance
        class. Caches it within the environment and returns the
        resulting ParsedTemplateResource.

        If force_reload is True, bypasses the cache.

        Note that this will also load any and all other templates that
        might be required (via nested slots) to render the passed
        template. These will be cached, but not directly returned.

        Note that cache operations here aren't threadsafe, but as long
        as the underlying resource doesn't change, the worst case is
        that we load and parse the same template resource multiple times
        without calling force_reload. This is probably better than
        wrapping the whole thing in a lock.

        Also note that we're prioritizing speed here in the cached happy
        case over technical correctness w.r.t. always raising if we
        don't support this loading flavor.
        """
        # Note: doing this first for thread safety in the sync case, preventing
        # any new template functions from being registered.
        self._has_loaded_any_template = True
        # Note: in this case, we can bypass iterally everything.
        if template is EmptyTemplate:
            return PARSED_EMPTY_TEMPLATE

        template_xable = cast(type[TemplateIntersectable], template)
        signature = template_xable._templatey_signature
        template_loader = self._get_loader(
            signature,
            self._can_load_sync,
            SyncTemplateLoader)

        with finalize_signature(
            signature,
            template,
            force_reload=force_reload,
            preload=preload,
            parse_cache=self._parsed_template_cache
        ) as sig_finalizer:
            for required_template_cls in sig_finalizer.required_loads:
                requirement_signature = cast(
                    type[TemplateIntersectable], required_template_cls
                )._templatey_signature
                requirement_text = template_loader.load_sync(
                    required_template_cls,
                    requirement_signature.resource_locator)
                parsed_requirement_template = self._parse_and_validate(
                    required_template_cls,
                    requirement_signature,
                    requirement_text,
                    override_validation_strictness)

                sig_finalizer.preload[required_template_cls] = \
                    parsed_requirement_template

        return sig_finalizer.target_resource

    def _get_loader[T: AsyncTemplateLoader | SyncTemplateLoader](
            self,
            signature: TemplateSignature,
            can_load_flavor: bool,
            flavor_cls: type[T]
            ) -> T:
        explicit_loader = signature.explicit_loader
        if explicit_loader is None:
            if not can_load_flavor:
                raise TypeError(
                    'Environment template loader does not support current '
                    + 'loading flavor (sync/async)', self._template_loader)

            # The cast here is because it could be a sync and/or async loader,
            # and the type system doesn't know we already verified that via
            # _can_load_<sync|async>.
            template_loader = cast(T, self._template_loader)
        else:
            if not isinstance(explicit_loader, flavor_cls):
                raise TypeError(
                    'Explicit template loader does not support current '
                    + 'loading flavor (sync/async)', explicit_loader)
            template_loader = explicit_loader

        return template_loader

    def _parse_and_validate(
            self,
            template_cls: TemplateClass,
            signature: TemplateSignature,
            template_text: str,
            override_validation_strictness: None | bool
            ) -> ParsedTemplateResource:
        parsed_template_resource = parse(
            template_text,
            signature.parse_config.interpolator,
            signature.parse_config.segment_modifiers)

        if override_validation_strictness is None:
            strict_mode = self.strict_interpolation_validation
        else:
            strict_mode = override_validation_strictness

        self._validate_env_functions(
            template_cls, parsed_template_resource)
        self._validate_template_signature(
            template_cls, signature, parsed_template_resource,
            strict_mode=strict_mode)

        return parsed_template_resource

    def _validate_env_functions(
            self,
            template_class: TemplateClass,
            parsed_template_resource: ParsedTemplateResource
            ) -> Literal[True]:
        """Makes sure that the template environment contains all of the
        template functions referenced in the template text. Returns
        True or raises MismatchedTemplateEnvironment.

        Note that we never use strict mode here, because it would be
        silly to require every single template to call every single
        template function. That's simply not what they're meant to be
        used for!
        """
        # Interestingly, .difference() works here, but plain ``-`` doesn't
        function_mismatch = parsed_template_resource.function_names.difference(
            self._env_functions)
        if function_mismatch:
            raise MismatchedTemplateEnvironment(
                'Template environment functions did not match the template '
                + 'text!', template_class, function_mismatch)

        for (
            function_name, function_calls
        ) in parsed_template_resource.function_calls.items():
            function_container = self._env_functions[function_name]
            for function_call in function_calls:
                try:
                    function_container.signature.bind(
                        *function_call.call_args,
                        **function_call.call_kwargs)
                except TypeError as exc:
                    raise MismatchedTemplateEnvironment(
                        'Template environment function had invalid call '
                        + 'signature', template_class, function_call
                    ) from exc

        return True

    def _validate_template_signature(
            self,
            template_class: TemplateClass,
            signature: TemplateSignature,
            parsed_template_resource: ParsedTemplateResource,
            *,
            strict_mode: bool
            ) -> Literal[True]:
        """Makes sure that the template signature includes all of the
        names referenced in the template text. Returns True or
        raises MismatchedTemplateSignature.
        """
        fieldset = signature.fieldset
        variable_names = fieldset.var_names
        slot_names = fieldset.slot_names
        dynamic_class_slot_names = fieldset.dynamic_class_slot_names
        content_names = fieldset.content_names
        data_names = fieldset.data_names

        if strict_mode:
            variables_mismatch = (
                parsed_template_resource.variable_names ^ variable_names)
            slot_mismatch = (
                parsed_template_resource.slot_names
                ^ (slot_names | dynamic_class_slot_names))
            content_mismatch = (
                parsed_template_resource.content_names ^ content_names)
            data_mismatch = (
                parsed_template_resource.data_names ^ data_names)

        else:
            variables_mismatch = (
                parsed_template_resource.variable_names - variable_names)
            slot_mismatch = (
                parsed_template_resource.slot_names
                - (slot_names | dynamic_class_slot_names))
            content_mismatch = (
                parsed_template_resource.content_names - content_names)
            data_mismatch = (
                parsed_template_resource.data_names - data_names)

        if (
            variables_mismatch
            or slot_mismatch
            or content_mismatch
            or data_mismatch
        ):
            raise MismatchedTemplateSignature(
                'Template interface variables, content, or slots did not '
                + 'match the template text!', template_class,
                variables_mismatch, slot_mismatch, content_mismatch,
                data_mismatch)

        return True

    def render_sync(
            self,
            template_instance: TemplateParamsInstance
            ) -> str:
        error_collector = ErrorCollector()
        template_preload: dict[TemplateClass, ParsedTemplateResource] = {}
        function_precall: dict[PrecallCacheKey, FuncExecutionResult] = {}

        render_ctx = RenderContext(
            template_preload=template_preload,
            function_precall=function_precall,
            error_collector=error_collector)
        for env_request in render_ctx.prep_render(template_instance):
            for root_to_load in env_request.to_load:
                # Note that this will already set the root_to_load within the
                # preload dict.
                self.load_sync(root_to_load, preload=template_preload)

            for to_execute in env_request.to_execute:
                self._execute_env_function_sync(
                    to_execute,
                    env_request.injections,
                    function_precall)

        to_join = render_driver(template_instance, render_ctx)
        if error_collector:
            raise ExceptionGroup('Failed to render template', error_collector)

        return ''.join(to_join)

    def _execute_env_function_sync(
            self,
            request: FuncExecutionRequest,
            injections: list[TemplateInjection],
            precall: dict[PrecallCacheKey, FuncExecutionResult]
            ) -> None:
        try:
            container = self._env_functions[request.name]
            if container.is_async:
                raise MismatchedRenderColor(
                    'Async env funcs cannot be used within render_sync!')

            exe_result = container.function(*request.args, **request.kwargs)
            precall[request.result_key] = FuncExecutionResult(
                name=request.name,
                retval=exe_result,
                exc=None)

            for result_segment in exe_result:
                if is_template_instance(result_segment):
                    injections.append((request.provenance, result_segment))

        except Exception as exc:
            precall[request.result_key] = FuncExecutionResult(
                name=request.name,
                retval=None,
                exc=exc)

    async def render_async(
            self,
            template_instance: TemplateParamsInstance
            ) -> str:
        error_collector = ErrorCollector()
        template_preload: dict[TemplateClass, ParsedTemplateResource] = {}
        function_precall: dict[PrecallCacheKey, FuncExecutionResult] = {}

        render_ctx = RenderContext(
            template_preload=template_preload,
            function_precall=function_precall,
            error_collector=error_collector)

        for env_request in render_ctx.prep_render(template_instance):
            async with create_task_group() as task_group:
                for root_to_load in env_request.to_load:
                    # Note that this will already set the root_to_load within
                    # the preload dict.
                    task_group.start_soon(partial(
                        self.load_async,
                        root_to_load,
                        preload=template_preload))

                for to_execute in env_request.to_execute:
                    task_group.start_soon(
                        self._execute_env_function_async,
                        to_execute,
                        env_request.injections,
                        function_precall)

        to_join = render_driver(template_instance, render_ctx)
        if error_collector:
            raise ExceptionGroup('Failed to render template', error_collector)

        return ''.join(to_join)

    async def _execute_env_function_async(
            self,
            request: FuncExecutionRequest,
            injections: list[TemplateInjection],
            precall: dict[PrecallCacheKey, FuncExecutionResult]
            ) -> None:
        try:
            container = self._env_functions[request.name]
            if container.is_async:
                exe_result = await container.function(
                    *request.args, **request.kwargs)
            else:
                exe_result = container.function(
                    *request.args, **request.kwargs)

            for result_segment in exe_result:
                if is_template_instance(result_segment):
                    injections.append((request.provenance, result_segment))

            precall[request.result_key] = FuncExecutionResult(
                name=request.name,
                retval=exe_result,
                exc=None)

        except Exception as exc:
            precall[request.result_key] = FuncExecutionResult(
                name=request.name,
                retval=None,
                exc=exc)


def _infer_asyncness(env_function: EnvFunction | EnvFunctionAsync) -> bool:
    """Infers, based on the type of the function, whether it should be
    considered sync or async.
    """
    return (
        inspect.iscoroutinefunction(env_function)
        or inspect.isawaitable(env_function))



from __future__ import annotations

import typing
from collections.abc import Sequence
from contextvars import ContextVar
from dataclasses import dataclass
from enum import Enum
from random import Random
from types import EllipsisType
from typing import Annotated
from typing import ClassVar
from typing import Literal
from typing import Protocol
from typing import TypeGuard

from docnote import ClcNote
from typing_extensions import TypeIs

if typing.TYPE_CHECKING:
    from _typeshed import DataclassInstance

    from templatey._signature import TemplateSignature
else:
    DataclassInstance = object


class InterfaceAnnotationFlavor(Enum):
    SLOT = 'slot'
    VARIABLE = 'var'
    CONTENT = 'content'
    DYNAMIC = 'dynamic'
    DATA = 'data'


@dataclass(frozen=True)
class InterfaceAnnotation:
    """
    """
    flavor: InterfaceAnnotationFlavor


class _TemplateClassSingleton(Enum):
    DYNAMIC_TEMPLATE_CLASS = 'dynamic'
type DynamicTemplateClass = Literal[
    _TemplateClassSingleton.DYNAMIC_TEMPLATE_CLASS]
DYNAMIC_TEMPLATE_CLASS: DynamicTemplateClass = \
    _TemplateClassSingleton.DYNAMIC_TEMPLATE_CLASS


# Technically this should be an intersection type with both the
# _TemplateIntersectable from templates and the DataclassInstance returned by
# the dataclass transform. Unfortunately, type intersections don't yet exist in
# python, so we have to resort to this (overly broad) type
TemplateParamsInstance = DataclassInstance
type TemplateClass = type[TemplateParamsInstance]
type TemplateInstanceID = int


# Technically, these should use the TemplateIntersectable from templates.py
# instead of ``TemplateParamsInstance``, python doesn't support type
# intersections yet, so we settle for this.
type Slot[T: TemplateParamsInstance] = Annotated[
    Sequence[T],
    InterfaceAnnotation(InterfaceAnnotationFlavor.SLOT),
    ClcNote(
        '''A ``Slot`` is a generic container used to define nested
        templates. They must be passed a concrete template class as a
        parameter, for example:

        > A simple template with a single slot
        __embed__: 'code/python'
            @template(my_template_config, my_template_locator)
            class EnclosingTemplate:
                my_slot: Slot[MyNestedTemplate]

        If you're looking for a way to define slots using a dynamic
        template class, you should either use ``DynamicSlot``
        (preferred) or look at the prebaked template injection function.
        ''')]
type Var[T] = Annotated[
    T | EllipsisType,
    InterfaceAnnotation(InterfaceAnnotationFlavor.VARIABLE),
    ClcNote(
        '''A ``Var`` is a generic container used to define some
        interpolatable variable within a template. They should be passed
        a concrete type as a parameter, though this value is not used
        by templatey. For example:

        > A simple template with a single ``Var``
        __embed__: 'code/python'
            @template(my_template_config, my_template_locator)
            class EnclosingTemplate:
                my_var: Var[str]

        The distinction between ``Content`` and ``Var`` is in how their
        values are handled. ``Content`` is intended only for **trusted**
        content; post-interpolation values are verified, but not
        escaped. Meanwhile, ``Var`` ^^can^^ be used with untrusted
        content, and its values will be escaped, but not verified.

        Both ``Var`` escaping and ``Content`` verification are
        controlled by the ``TemplateConfig``.

        Note that ``Var`` values within templates may be provided by
        an enclosing template, as a parameter on their slots. For more
        information, <create a github issue if you read this please,
        because we haven't made a guide for this yet>
        ''')]
type Content[T] = Annotated[
    T,
    InterfaceAnnotation(InterfaceAnnotationFlavor.CONTENT),
    ClcNote(
        '''A ``Content`` parameter is a generic container used to define some
        interpolatable variable within a template. They should be passed
        a concrete type as a parameter, though this value is not used
        by templatey. For example:

        > A simple template with a single ``Content`` param
        __embed__: 'code/python'
            @template(my_template_config, my_template_locator)
            class EnclosingTemplate:
                my_var: Content[str]

        The distinction between ``Content`` and ``Var`` is in how their
        values are handled. ``Content`` is intended only for **trusted**
        content; post-interpolation values are verified, but not
        escaped. Meanwhile, ``Var`` ^^can^^ be used with untrusted
        content, and its values will be escaped, but not verified.

        Both ``Var`` escaping and ``Content`` verification are
        controlled by the ``TemplateConfig``.
        ''')]
type DynamicClassSlot[T: TemplateParamsInstance] = Annotated[
    Sequence[T],
    InterfaceAnnotation(InterfaceAnnotationFlavor.SLOT),
    InterfaceAnnotation(InterfaceAnnotationFlavor.DYNAMIC),
    ClcNote(
        '''Like a ``Slot``, a ``DynamicClassSlot`` is a generic container
        used to define nested templates. Although templates in
        templatey cannot be subclassed, they may use mixins as
        superclasses; in that case, it may be useful to supply a type
        argument to the dynamic slot to narrow its type during type
        checking.

        > A simple template with two dynamic slots
        __embed__: 'code/python'
            @template(my_template_config, my_template_locator)
            class EnclosingTemplate:
                # Note that this can be literally any template instance
                my_slot_1: DynamicClassSlot
                # This is constrained to a subclass
                my_slot_2: DynamicClassSlot[SomeMixin]

        Note that the type arg to ``DynamicClassSlot`` is neither required,
        nor used by templatey. It's provided strictly to support
        applications where you want more restrictive type annotations
        than "literally any template class whatsoever".

        Note also that ``DynamicClassSlot``s come at a slight performance
        penalty compared to normal ``Slot``s. If possible, prefer a
        plain ``Slot``, perhaps with a type union as a parameter.
        That being said, ``DynamicClassSlot``s are still more performant than
        injecting templates via environment functions.
        ''')]


def is_template_class_xable(cls: type) -> TypeIs[type[TemplateIntersectable]]:
    """Rather than relying upon @runtime_checkable, which doesn't work
    with protocols with ClassVars, we implement our own custom checker
    here for narrowing the type against TemplateIntersectable. Note
    that this also, I think, might be usable for some of the issues
    re: the missing intersection type in python, though support might be
    unreliable depending on which type checker is in use.
    """
    return hasattr(cls, '_templatey_signature')


def is_template_instance_xable(
        obj: object
        ) -> TypeIs[TemplateIntersectable]:
    """Rather than relying upon @runtime_checkable, which doesn't work
    with protocols with ClassVars, we implement our own custom checker
    here for narrowing the type against TemplateIntersectable. Note
    that this also, I think, might be usable for some of the issues
    re: the missing intersection type in python, though support might be
    unreliable depending on which type checker is in use.
    """
    return is_template_class_xable(type(obj))


def is_template_class(cls: type) -> TypeGuard[TemplateClass]:
    """Like is_template_class_xable, but for the raw template class.
    """
    return isinstance(cls, type) and hasattr(cls, '_templatey_signature')


def is_template_instance(obj: object) -> TypeGuard[TemplateParamsInstance]:
    """Like is_template_instance_xable, but for the raw template class.
    """
    return (not isinstance(obj, type)) and hasattr(obj, '_templatey_signature')


class TemplateIntersectable(Protocol):
    """This is the actual template protocol, which we would
    like to intersect with the TemplateParamsInstance, but cannot
    (because python currently lacks an intersection type).

    Partly here for documentation, partly for use in ``cast`` calls.
    """
    _templatey_signature: ClassVar[TemplateSignature]


# Note: we don't need cryptographically secure IDs here, so let's preserve
# entropy (might also be faster, dunno). Also note: the only reason we're
# using a contextvar here is so that we can theoretically replace it with
# a deterministic seed during testing (if we run into flakiness due to
# non-determinism)
_ID_PRNG: ContextVar[Random] = ContextVar('_ID_PRNG', default=Random())  # noqa: B039, S311
_ID_BITS = 128


def create_templatey_id() -> int:
    """Templatey IDs are unique identifiers (theoretically, absent
    birthday collisions) that we currently use in two places:
    ++  as a scope ID, which is used when defining templates in closures
    ++  for giving slot/prerender tree nodes a unique reference target for
        recursion loops while copying and merging, which is more robust
        than ``id(target)`` and can be transferred via dataclass field
        into cloned/copied/merged nodes.
    """
    prng = _ID_PRNG.get()
    return prng.getrandbits(_ID_BITS)



from __future__ import annotations

import typing
from collections import namedtuple
from collections.abc import Iterator
from dataclasses import Field
from dataclasses import dataclass
from dataclasses import field
from dataclasses import fields
from textwrap import dedent
from types import UnionType
from typing import Any
from typing import NamedTuple
from typing import TypeAliasType
from typing import cast
from typing import get_args as get_type_args
from typing import get_origin as get_type_origin
from typing import get_type_hints

from templatey._types import Content
from templatey._types import DynamicClassSlot
from templatey._types import InterfaceAnnotationFlavor
from templatey._types import Slot
from templatey._types import TemplateClass
from templatey._types import Var
from templatey.templates import FieldConfig
from templatey.templates import get_closure_locals

if typing.TYPE_CHECKING:
    from templatey._slot_tree import SlotPath

type SlotFieldAnnotation = (
    TemplateClass
    | UnionType
    | TypeAliasType)


@dataclass(slots=True, frozen=True, kw_only=True)
class NormalizedField:
    name: str
    config: FieldConfig


@dataclass(slots=True, frozen=True, kw_only=True)
class NormalizedSlotField(NormalizedField):
    type_: SlotFieldAnnotation


@dataclass(slots=True)
class NormalizedFieldset:
    slots: tuple[NormalizedSlotField, ...]
    dynamic_slots: tuple[NormalizedField, ...]
    vars_: tuple[NormalizedField, ...]
    content: tuple[NormalizedField, ...]
    data: tuple[NormalizedField, ...]

    data_names: frozenset[str] = field(init=False)
    var_names: frozenset[str] = field(init=False)
    content_names: frozenset[str] = field(init=False)
    # Note that these are all ONLY the direct nesteds; these do not include
    # anything from deeper in the slot tree.
    slot_names: frozenset[str] = field(init=False)
    dynamic_class_slot_names: frozenset[str] = field(init=False)
    slotpaths: frozenset[SlotPath]

    # Oldschool here for performance reasons; otherwise this would be a dict.
    # Field names match the field names from the params; the value is gathered
    # from the metadata value on the field.
    transformers: NamedTuple

    def __post_init__(self):
        self.data_names = frozenset(field.name for field in self.data)
        self.var_names = frozenset(field.name for field in self.vars_)
        self.content_names = frozenset(field.name for field in self.content)
        self.slot_names = frozenset(field.name for field in self.slots)
        self.dynamic_class_slot_names = frozenset(
            field.name for field in self.dynamic_slots)

    # The noqa here is because of too many branches (because we're normalizing
    # stuff, basically)
    @classmethod
    def from_template_cls(  # noqa: PLR0912
            cls,
            template_cls: TemplateClass
            ) -> NormalizedFieldset:
        closure_locals = get_closure_locals(template_cls)
        try:
            if closure_locals is None:
                template_type_hints = get_type_hints(template_cls)
            else:
                template_type_hints = get_type_hints(
                    template_cls, localns=closure_locals)
        except NameError as exc:
            exc.add_note(dedent('''\
                Failed to resolve type hints on template. This is either the
                result of an unresolved forward ref (which should have been
                caught by your type checker), or circular imports hidden behind
                ``if typing.TYPE_CHECKING`` blocks. These must be resolved
                before calculating a template signature.'''))
            raise exc

        slotpaths: set[SlotPath] = set()
        slots = []
        dynamic_slots = []
        vars_ = []
        content = []
        data = []
        transformers = {}
        # Note that this ignores initvars, which is what we want
        for template_field in fields(template_cls):
            field_classification = _classify_interface_field_flavor(
                template_type_hints, template_field)

            # Note: it's not entirely clear to me that this restriction makes
            # sense; I could potentially see MAYBE there being some kind of
            # environment function that could access other attributes from the
            # dataclass? But also, maybe those should be vars? Again, unclear.
            if field_classification is None:
                raise TypeError(
                    '``template_field`` definitions may only contain '
                    + 'variables, slots, and content!')

            else:
                field_flavors, wrapped_type = field_classification

                # In case it isn't obvious, the .get() here is because we don't
                # require all fields to use the ``template_field`` specifier.
                field_config: FieldConfig | None = template_field.metadata.get(
                    'templatey.field_config')
                if field_config is None:
                    field_config = FieldConfig()

                # Just for maintainability, we're doing this first here and
                # then replacing it for slots; that way, we don't have a bunch
                # of these
                normfield: NormalizedField | NormalizedSlotField = \
                    NormalizedField(
                        name=template_field.name,
                        config=field_config)

                # A little awkward to effectively just repeat the comparison we
                # did when classifying, but that makes testing easier and
                # control flow clearer
                if InterfaceAnnotationFlavor.VARIABLE in field_flavors:
                    vars_.append(normfield)

                elif InterfaceAnnotationFlavor.SLOT in field_flavors:
                    if InterfaceAnnotationFlavor.DYNAMIC in field_flavors:
                        dynamic_slots.append(normfield)

                    else:
                        slot_annotation = cast(
                            SlotFieldAnnotation, wrapped_type)
                        normfield = NormalizedSlotField(
                            name=normfield.name,
                            config=normfield.config,
                            type_=slot_annotation)
                        slots.append(normfield)
                        # Just to be clear: we're flattening unions etc here.
                        slotpaths.update(
                            (template_field.name, normtype)
                            for normtype in normalize_slot_annotations(
                                slot_annotation))

                elif InterfaceAnnotationFlavor.CONTENT in field_flavors:
                    content.append(normfield)

                else:
                    data.append(normfield)

                transformers[template_field.name] = \
                    normfield.config.transformer

        transformer_cls = namedtuple(
            'Templateytransformers', tuple(transformers))

        return NormalizedFieldset(
            slots=tuple(slots),
            dynamic_slots=tuple(dynamic_slots),
            vars_=tuple(vars_),
            content=tuple(content),
            data=tuple(data),
            transformers=transformer_cls(**transformers),
            slotpaths=frozenset(slotpaths))


# Yes, this is awkward with a bazillion return statements, but in this case,
# clarity is better than elegance
def _classify_interface_field_flavor(  # noqa: PLR0911
        parent_class_type_hints: dict[str, Any],
        template_field: Field
        ) -> tuple[set[InterfaceAnnotationFlavor], type | None]:
    """For a dataclass field, determines whether it was declared as a
    var, slot, or content.

    If none of the above, returns None.
    """
    # Note that dataclasses don't include the actual type (just a string)
    # when in __future__ mode, so we need to get them from the parent class
    # by calling get_type_hints() on it
    resolved_field_type = parent_class_type_hints[template_field.name]
    anno_origin = get_type_origin(resolved_field_type)
    if anno_origin is Var:
        nested_type, = get_type_args(resolved_field_type)
        return {InterfaceAnnotationFlavor.VARIABLE}, nested_type
    elif anno_origin is Slot:
        nested_type, = get_type_args(resolved_field_type)
        return {InterfaceAnnotationFlavor.SLOT}, nested_type
    elif anno_origin is Content:
        nested_type, = get_type_args(resolved_field_type)
        return {InterfaceAnnotationFlavor.CONTENT}, nested_type
    elif anno_origin is DynamicClassSlot:
        return {
            InterfaceAnnotationFlavor.SLOT,
            InterfaceAnnotationFlavor.DYNAMIC
        }, None

    # This is all if there's a generic annotation with no parameter passed,
    # ex ``foo: Var`` or ``bar: Slot`` or (presumably) ``baz: list``
    elif anno_origin is None:
        if resolved_field_type is Var:
            return {InterfaceAnnotationFlavor.VARIABLE}, None

        elif resolved_field_type is Content:
            return {InterfaceAnnotationFlavor.CONTENT}, None

        elif resolved_field_type is DynamicClassSlot:
            return {
                InterfaceAnnotationFlavor.SLOT,
                InterfaceAnnotationFlavor.DYNAMIC
            }, None

        elif resolved_field_type is Slot:
            raise TypeError(
                '``Slot`` annotations require a concrete slot class as a '
                + 'parameter!')

        else:
            return {InterfaceAnnotationFlavor.DATA}, None

    # There was a non-generic parameter passed. Therefore, it can only be
    # one thing: template data.
    else:
        return {InterfaceAnnotationFlavor.DATA}, None


def normalize_slot_annotations(
        slot_annotation: SlotFieldAnnotation
        ) -> Iterator[TemplateClass]:
    # Note that this still might contain a heterogeneous mix of
    # template classes and forward refs! Hence flattening first.
    if isinstance(slot_annotation, UnionType):
        for union_member in slot_annotation.__args__:
            yield from normalize_slot_annotations(union_member)

    elif isinstance(slot_annotation, TypeAliasType):
        yield from normalize_slot_annotations(
            slot_annotation.__value__)

    else:
        yield slot_annotation



from __future__ import annotations

import ast
import itertools
import logging
import re
import string
from collections import defaultdict
from collections.abc import Collection
from collections.abc import Generator
from collections.abc import Iterator
from collections.abc import Mapping
from collections.abc import Sequence
from dataclasses import dataclass
from dataclasses import field
from dataclasses import fields
from dataclasses import replace as dc_replace
from functools import singledispatch
from itertools import count
from typing import Annotated
from typing import Any
from typing import cast

from docnote import ClcNote

from templatey.exceptions import DuplicateSlotName
from templatey.exceptions import InvalidTemplateInterpolation
from templatey.interpolators import NamedInterpolator
from templatey.interpolators import transform_unicode_control
from templatey.interpolators import untransform_unicode_control
from templatey.modifiers import EnvFuncInvocationRef
from templatey.modifiers import SegmentModifier

_SLOT_MATCHER = re.compile(r'^\s*slot\.([A-z_][A-z0-9_]*)\s*$')
_CONTENT_MATCHER = re.compile(r'^\s*content\.([A-z_][A-z0-9_]*)\s*$')
_VAR_MATCHER = re.compile(r'^\s*var\.([A-z_][A-z0-9_]*)\s*$')
_FUNC_MATCHER = re.compile(r'^\s*@([A-z_][A-z0-9_]*)\(([^\)]*)\)\s*$')
_COMMENT_MATCHER = re.compile(r'^\s*#.*$')
logger = logging.getLogger(__name__)

type TemplateSegment = (
    LiteralTemplateString
    | InterpolatedSlot
    | InterpolatedContent
    | InterpolatedVariable
    | InterpolatedFunctionCall)


@dataclass(frozen=True, slots=True)
class ParsedTemplateResource:
    """In addition to storing the actual template parts, this stores
    information about which references the template had, for use later
    when validating the template (within some render context).
    """
    parts: tuple[TemplateSegment, ...]
    variable_names: frozenset[str]
    content_names: frozenset[str]
    slot_names: frozenset[str]
    function_names: frozenset[str]
    data_names: frozenset[str]
    # Separate this out from function_names so that we can put compare=False
    # while still preserving comparability between instances. It's not clear if
    # this is useful, but the memory footprint should be low
    # Note: this is included for convenience, so that the render environment
    # has easy access to all of the *args and **kwargs, so that they can be
    # tested against the signature of the actual render function during loading
    function_calls: dict[str, tuple[InterpolatedFunctionCall, ...]] = field(
        compare=False)
    slots: dict[str, InterpolatedSlot] = field(compare=False)

    # Purely here for performance reasons
    part_count: int = field(init=False, compare=False)

    def __post_init__(self):
        object.__setattr__(self, 'part_count', len(self.parts))

    @classmethod
    def from_parts(
            cls,
            parts: Sequence[TemplateSegment]
            ) -> ParsedTemplateResource:
        if not isinstance(parts, tuple):
            parts = tuple(parts)

        content_names = set()
        slot_names = set()
        variable_names = set()
        data_names = set()
        functions = defaultdict(list)
        for part in parts:
            if isinstance(part, InterpolatedContent):
                content_names.add(part.name)
            elif isinstance(part, InterpolatedVariable):
                variable_names.add(part.name)
            elif isinstance(part, InterpolatedSlot):
                # Interpolated slots must be unique; enforce that here.
                if part.name in slot_names:
                    raise DuplicateSlotName(part.name)

                for maybe_reference in part.params.values():
                    nested_content_refs, nested_var_refs, _ = \
                        _extract_nested_refs(maybe_reference)
                    content_names.update(
                        ref.name for ref in nested_content_refs)
                    variable_names.update(ref.name for ref in nested_var_refs)

                slot_names.add(part.name)

            elif isinstance(part, InterpolatedFunctionCall):
                for maybe_reference in itertools.chain(
                    part.call_args,
                    part.call_kwargs.values(),
                    (starargs for starargs in (part.call_args_exp,)),
                    (starkwargs for starkwargs in (part.call_kwargs_exp,))
                ):
                    (
                        nested_content_refs,
                        nested_var_refs,
                        nested_data_refs) = _extract_nested_refs(
                            maybe_reference)
                    content_names.update(
                        ref.name for ref in nested_content_refs)
                    variable_names.update(ref.name for ref in nested_var_refs)
                    data_names.update(ref.name for ref in nested_data_refs)

                functions[part.name].append(part)

        return cls(
            parts=parts,
            content_names=frozenset(content_names),
            variable_names=frozenset(variable_names),
            slot_names=frozenset(slot_names),
            function_names=frozenset(functions),
            function_calls={
                name: tuple(calls) for name, calls in functions.items()},
            data_names=frozenset(data_names),
            slots={
                maybe_slot.name: maybe_slot
                for maybe_slot in parts
                if isinstance(maybe_slot, InterpolatedSlot)})


class LiteralTemplateString(str):
    __slots__ = ['part_index']
    part_index: int

    def __new__(cls, *args, part_index: int, **kwargs):
        instance = super().__new__(cls, *args, **kwargs)
        instance.part_index = part_index
        return instance


@dataclass(frozen=True, slots=True)
class InterpolatedContent:
    """
    """
    part_index: int
    # TODO: this needs a way to define any variables used by the content via
    # ComplexContent! Otherwise, strict mode in template/interface validation
    # will always fail with interpolated content.
    name: str
    config: InterpolationConfig


@dataclass(frozen=True, slots=True)
class InterpolatedSlot:
    """
    """
    part_index: int
    name: str
    params: dict[str, object]
    config: InterpolationConfig

    def __post_init__(self):
        if self.config.fmt:
            raise InvalidTemplateInterpolation(
                'Template slots cannot have format specs!')


@dataclass(frozen=True, slots=True)
class InterpolatedVariable:
    """
    """
    part_index: int
    name: str
    config: InterpolationConfig

    def __post_init__(self):
        if self.config.prefix is not None or self.config.suffix is not None:
            raise InvalidTemplateInterpolation(
                'Template variables cannot have prefixes nor suffixes. If you '
                + 'need similar behavior, you can create a dedicated '
                + 'converter function to add the affix before escaping.')


@dataclass(frozen=True, slots=True)
class InterpolatedFunctionCall:
    """
    """
    part_index: int
    name: str
    call_args: Sequence[object] = field(compare=False)
    call_args_exp: object | None = field(compare=False)
    call_kwargs: dict[str, object] = field(compare=False)
    call_kwargs_exp: object | None = field(compare=False)

    def _matches(self, other: object) -> bool:
        """Okay, so the problem is as follows:
        1.. We need to be able to check equality during testing
        2.. We need hashing in other places
        3.. We can't mess with equality, because it would break the
            dict keys we need from the hash
        4.. Therefore, we need to exempt the call args / etc from the
            normal equality check, but still need something for testing
        Therefore, we create this, which checks everything -- but just
        for testing.
        """
        if not isinstance(other, InterpolatedFunctionCall):
            return False

        return (
            self.part_index == other.part_index
            and self.name == other.name
            and self.call_args == other.call_args
            and self.call_args_exp == other.call_args_exp
            and self.call_kwargs == other.call_kwargs
            and self.call_kwargs_exp == other.call_kwargs_exp)


@dataclass(slots=True, frozen=True)
class TemplateInstanceContentRef:
    """Used to indicate that an environment function or segment
    modification needs to reference a content parameter on the current
    template instance being rendered.
    """
    name: Annotated[
        str,
        ClcNote('The name of the content parameter')]


@dataclass(slots=True, frozen=True)
class TemplateInstanceVariableRef:
    """Used to indicate that an environment function or segment
    modification needs to reference a variable parameter on the current
    template instance being rendered.
    """
    name: Annotated[
        str,
        ClcNote('The name of the variable parameter')]


@dataclass(slots=True, frozen=True)
class TemplateInstanceDataRef:
    """Used to indicate that an environment function (including one
    injected via segment modification) needs to reference a template
    data attribute on the current template instance being rendered.
    """
    name: Annotated[
        str,
        ClcNote('The name of the data attribute (the dataclass field name)')]


_VALID_NESTED_REFS = {
    'content': TemplateInstanceContentRef,
    'var': TemplateInstanceVariableRef,
    'data': TemplateInstanceDataRef,}


def parse(
        template_str: str,
        interpolator: NamedInterpolator,
        segment_modifiers: Sequence[SegmentModifier],
        ) -> ParsedTemplateResource:
    """Parses the template string with the passed interpolator and
    applies the desired segment modifiers.
    """
    if interpolator is NamedInterpolator.UNICODE_CONTROL:
        do_untransform = True
        template_str = transform_unicode_control(template_str)
    else:
        do_untransform = False

    part_index_counter = count()
    parts: list[TemplateSegment] = []
    for premodification_segment in _wrap_formatter_parse(
        template_str, do_untransform=do_untransform
    ):
        parts.extend(_apply_segment_modifiers(
            premodification_segment,
            segment_modifiers,
            part_index_counter))
    return ParsedTemplateResource.from_parts(parts)


def _apply_segment_modifiers(
        unmodified_part: TemplateSegment,
        segment_modifiers: Sequence[SegmentModifier],
        part_index_counter: count,
        ) -> Iterator[TemplateSegment]:
    """As the name suggests, this takes an unmodified part and applies
    any desired segment modifiers. Note that this can grow or contract
    the total number of segments.
    """
    # The combinatorics here are gross, but this only runs once per
    # template load, and not once per render, so at least there's
    # that.
    if isinstance(unmodified_part, str):
        for modifier in segment_modifiers:
            had_matches, after_mods = modifier.apply_and_flatten(
                unmodified_part)

            if had_matches:
                yield from (
                    _coerce_modified_segment(
                        modified_segment, part_index_counter)
                    for modified_segment in after_mods)
                break

        else:
            # We still want to create a copy here, in case the
            # template loader is doing its own caching beyond what
            # we do within the env
            yield LiteralTemplateString(
                unmodified_part,
                part_index=next(part_index_counter))

    else:
        yield dc_replace(
            unmodified_part,
            part_index=next(part_index_counter))


def _coerce_modified_segment(
        modified_segment:
            str
            | EnvFuncInvocationRef
            | TemplateInstanceContentRef
            | TemplateInstanceVariableRef,
        part_index_counter: count,
            ) -> (
                LiteralTemplateString
                | InterpolatedContent
                | InterpolatedVariable
                | InterpolatedFunctionCall):
    """Converts the result of segment modification into an actual
    parsed template resource part, including a part index.
    """
    if isinstance(modified_segment, str):
        return LiteralTemplateString(
            modified_segment, part_index=next(part_index_counter))

    elif isinstance(modified_segment, EnvFuncInvocationRef):
        return InterpolatedFunctionCall(
            part_index=next(part_index_counter),
            name=modified_segment.name,
            call_args=modified_segment.call_args,
            call_args_exp=None,
            call_kwargs=modified_segment.call_kwargs,
            call_kwargs_exp=None)

    elif isinstance(modified_segment, TemplateInstanceContentRef):
        return InterpolatedContent(
            part_index=next(part_index_counter),
            name=modified_segment.name,
            config=InterpolationConfig())

    elif isinstance(modified_segment, TemplateInstanceVariableRef):
        return InterpolatedVariable(
            part_index=next(part_index_counter),
            name=modified_segment.name,
            config=InterpolationConfig())

    else:
        raise TypeError('Unknown modified segment type!', modified_segment)


def _extract_nested_refs(
        value
        ) -> tuple[
            set[TemplateInstanceContentRef],
            set[TemplateInstanceVariableRef],
            set[TemplateInstanceDataRef]]:
    """Call this to recursively extract all of the content and variable
    references contained within an environment function call.
    """
    content_refs = set()
    var_refs = set()
    data_refs = set()

    # Note that order here is important! Mappings are always collections!
    # Strings are always collections, too!
    if isinstance(value, Mapping):
        nested_values = value.values()
    elif isinstance(value, str):
        nested_values = ()
    elif isinstance(value, Collection):
        nested_values = value
    else:
        nested_values = ()

        if isinstance(value, TemplateInstanceContentRef):
            content_refs.add(value)

        elif isinstance(value, TemplateInstanceVariableRef):
            var_refs.add(value)

        elif isinstance(value, TemplateInstanceDataRef):
            data_refs.add(value)

    for nested_val in nested_values:
        nested_content_refs, nested_var_refs, nested_data_refs = \
            _extract_nested_refs(nested_val)
        content_refs.update(nested_content_refs)
        var_refs.update(nested_var_refs)
        data_refs.update(nested_data_refs)

    return content_refs, var_refs, data_refs


def _wrap_formatter_parse(
        formattable_template_str: str,
        do_untransform=False
        ) -> Generator[TemplateSegment, None, None]:
    """A generator. Wraps the very weird API provided by
    string.Formatter.parse, instead yielding either:
    ++  literal text, in string format
    ++  ``InterpolatedContent`` instances
    ++  ``InterpolatedSlot`` instances
    ++  ``InterpolatedVariable`` instances
    ++  ``InterpolatedFunctionCall`` instances
    """
    part_counter = itertools.count()
    formatter = string.Formatter()

    for format_tuple in formatter.parse(formattable_template_str):
        # Arg order here is: literal_text, field_name, format_spec, conversion
        # Note that this can contain BOTH a literal text and a field name.
        # It's a really weird API; that's why we're wrapping it in
        # _extract_formatting_kwargs. It still reads things left-to-right, but
        # it bundles them together really strangely
        if do_untransform:
            literal_text, field_name, format_spec, conversion = (
                untransform_unicode_control(format_tuple_part)
                if format_tuple_part is not None else None
                for format_tuple_part in format_tuple)
        else:
            literal_text, field_name, format_spec, conversion = format_tuple

        if literal_text is not None:
            if do_untransform:
                yield LiteralTemplateString(
                    untransform_unicode_control(literal_text),
                    part_index=next(part_counter))
            else:
                yield LiteralTemplateString(
                    literal_text,
                    part_index=next(part_counter))

        # field_name can be None, an empty string, or a kwargname.
        # None means there's no formatting field left in the string -- in which
        # case, the literal_text would contain the rest of the string.
        if field_name is None:
            continue
        else:
            coerced_interpolation = _coerce_interpolation(
                field_name, format_spec, conversion, part_counter)
            if coerced_interpolation is not None:
                yield coerced_interpolation


@dataclass(frozen=True, slots=True)
class InterpolationConfig:
    """
    """
    fmt: str | None = None
    prefix: str | None = None
    suffix: str | None = None

    def apply_prefix_iter(self) -> Iterator[str]:
        """If a prefix is defined, yields it. Otherwise, does nothing.
        """
        prefix = self.prefix
        if prefix is not None:
            yield prefix

    def apply_suffix_iter(self) -> Iterator[str]:
        """If a suffix is defined, yields it. Otherwise, does nothing.
        """
        suffix = self.suffix
        if suffix is not None:
            yield suffix

    def apply_affix(self, val: str | None) -> tuple[str, ...]:
        """For the given val, inserts any defined prefix and/or suffix.
        If val is None, returns an empty tuple.

        Note that tuples are faster for list.extend than both iterators
        and lists, at least for these small sizes.
        """
        if val is None:
            return ()

        suffix = self.suffix
        prefix = self.prefix
        # Explicitly typing out the logic square here as a microoptimization
        if prefix is None:
            if suffix is None:
                return (val,)
            else:
                return (val, suffix)
        elif suffix is None:
            return (prefix, val)
        else:
            return (prefix, val, suffix)

    @classmethod
    def from_format_spec(
            cls,
            format_spec: str | None
            ) -> tuple[InterpolationConfig, dict[str, object]]:
        if format_spec is None or not format_spec:
            return (cls(), {})

        # If you shift away from using AST-based parsing, then this trick is
        # useful for converting raw strings to normal ones:
        # codecs.decode(r'\n', 'unicode_escape') == '\n'
        try:
            tree = ast.parse(f'print({format_spec})')
        except (ValueError, SyntaxError) as exc:
            raise InvalidTemplateInterpolation(
                'Invalid interpolation parameters!'
            ) from exc

        injected_print = cast(ast.Call, cast(ast.Expr, tree.body[0]).value)

        kwargs = {}
        dunders = {}

        if injected_print.args:
            raise InvalidTemplateInterpolation(
                'Everything after the : in a templatey interpolation must be '
                + 'a keyword-only argument dict!')

        for ast_kwarg in injected_print.keywords:
            argname = ast_kwarg.arg
            if argname is None:
                raise InvalidTemplateInterpolation(
                    'Additional arguments in templatey interpolations (ex '
                    + 'explicit slot parameters) do not currently support '
                    + 'star expansion (**kwargs)')

            elif argname.startswith('__') and argname.endswith('__'):
                dunder_name = argname[2:-2]

                if dunder_name not in _interp_cfg_fields:
                    logger.warning('Skipping unknown dunder field %s', argname)
                    continue

                # I think the following bit is reporting as unreachable because
                # it doesn't think the singledispatch will match on anything,
                # and therefore raise valueerror
                dunder_value = _extract_reference_or_literal(ast_kwarg.value)

                if not isinstance(dunder_value, str):
                    raise InvalidTemplateInterpolation(
                        'Non-string values for interpolation dunders are not '
                        + 'currently supported!')

                dunders[dunder_name] = dunder_value

            else:
                kwargs[ast_kwarg.arg] = _extract_reference_or_literal(
                    ast_kwarg.value)

        return (cls(**dunders), kwargs)


# We use this for faster checks when parsing templates in from_format_spec
_interp_cfg_fields = {field.name for field in fields(InterpolationConfig)}


def _coerce_interpolation(field_name, format_spec, conversion, part_counter):
    try:
        if conversion:
            raise InvalidTemplateInterpolation(
                'Conversion specs are not supported in templatey; pass a '
                + 'stringifier to the field definition instead.')

        # The format spec is determined by the first : in the interpolation.
        # Any following :s are included as part of it. However, for function
        # calls, we want to interpret the format spec : as literally part of
        # the field_name, so we join them back up.
        if format_spec:
            full_interpolation_def = f'{field_name}:{format_spec}'
        else:
            full_interpolation_def = field_name

        if _COMMENT_MATCHER.match(full_interpolation_def) is not None:
            return None

        if (match := _FUNC_MATCHER.match(full_interpolation_def)) is not None:
            args, starargs, kwargs, starkwargs = _extract_call_signature(
                match.group(2))
            return InterpolatedFunctionCall(
                part_index=next(part_counter),
                name=match.group(1),
                call_args=args,
                call_args_exp=starargs,
                call_kwargs=kwargs,
                call_kwargs_exp=starkwargs,)

        interp_cfg, kwargs = InterpolationConfig.from_format_spec(format_spec)
        if (match := _SLOT_MATCHER.match(field_name)) is not None:
            slot_params_str = format_spec.strip()
            args, starargs, kwargs, starkwargs = _extract_call_signature(
                slot_params_str)

            return InterpolatedSlot(
                part_index=next(part_counter),
                name=match.group(1),
                params=kwargs,
                config=interp_cfg)

        else:
            if kwargs:
                raise InvalidTemplateInterpolation(
                    'Interpolated variables and content cannot have arbitrary '
                    + 'kwargs, only dunders!')

            if (match := _VAR_MATCHER.match(field_name)) is not None:
                return InterpolatedVariable(
                    part_index=next(part_counter),
                    name=match.group(1),
                    config=interp_cfg)

            elif (match := _CONTENT_MATCHER.match(field_name)) is not None:
                return InterpolatedContent(
                    part_index=next(part_counter),
                    name=match.group(1),
                    config=interp_cfg)

        raise InvalidTemplateInterpolation(
            'Unknown target for templatey interpolation (must be var, slot, '
            + 'env function, content, or comment)')

    except InvalidTemplateInterpolation as exc:
        exc.add_note(f'{field_name=}, {format_spec=}, {conversion=}')
        raise exc


def _extract_call_signature(str_signature):
    """Returns *args and **kwargs for the desired asset function."""
    try:
        tree = ast.parse(f'print({str_signature})')
        injected_print = cast(ast.Call, cast(ast.Expr, tree.body[0]).value)

        args = []
        starargs = None
        kwargs = {}
        starkwargs = None

        for ast_arg in injected_print.args:
            if isinstance(ast_arg, ast.Starred):
                starargs = _extract_reference_or_literal(ast_arg.value)
            else:
                args.append(_extract_reference_or_literal(ast_arg))

        for ast_kwarg in injected_print.keywords:
            if ast_kwarg.arg is None:
                starkwargs = _extract_reference_or_literal(ast_kwarg.value)
            else:
                kwargs[ast_kwarg.arg] = _extract_reference_or_literal(
                    ast_kwarg.value)

        return args, starargs, kwargs, starkwargs

    except (ValueError, SyntaxError) as exc:
        raise InvalidTemplateInterpolation(
            'Invalid environment function call signature') from exc


@singledispatch
def _extract_reference_or_literal(ast_node) -> Any:
    """Gets the actual reference out of an AST node used in the call
    signature, either as an arg or the value of a kwarg.
    """
    raise ValueError('No matching node type', ast_node)


@_extract_reference_or_literal.register
def _(ast_node: ast.Attribute):
    should_be_name = ast_node.value
    if not isinstance(should_be_name, ast.Name):
        raise ValueError('Non-literals must be attributes', ast_node)

    target_cls = _VALID_NESTED_REFS.get(should_be_name.id)
    if target_cls is None:
        raise ValueError(
            'Invalid asset reference value for asset function',
            should_be_name.id)

    return target_cls(name=ast_node.attr)


@_extract_reference_or_literal.register
def _(ast_node: ast.Constant):
    return ast_node.value



"""These are responsible for finalizing the template definition during
loading. We defer until loading so that forward refs on the templates
are resolved. Trust me, it's way, WAY easier this way (we used to try
to do it the other way and it was a massive headache.
"""
from __future__ import annotations

from collections.abc import Generator
from contextlib import contextmanager
from dataclasses import dataclass
from dataclasses import field
from typing import cast

from templatey._fields import NormalizedFieldset
from templatey._signature import TemplateSignature
from templatey._slot_tree import build_slot_tree
from templatey._types import TemplateClass
from templatey._types import TemplateIntersectable
from templatey.parser import ParsedTemplateResource


@dataclass(slots=True)
class SignatureFinalizer:
    preload: dict[TemplateClass, ParsedTemplateResource]
    required_loads: set[TemplateClass]
    target_resource: ParsedTemplateResource = field(init=False)


@contextmanager
def finalize_signature(
        signature: TemplateSignature,
        template_cls: TemplateClass,
        *,
        force_reload: bool = False,
        preload: dict[TemplateClass, ParsedTemplateResource] | None,
        parse_cache: dict[TemplateClass, ParsedTemplateResource],
        ) -> Generator[SignatureFinalizer, None, None]:
    """This contextmanager centralizes all the logic to finalize the
    signature for a template during the loading process.
    """
    # As the name suggests, this also recursively finalizes all inclusions.
    ensure_recursive_totality(signature, template_cls)

    # Note that, because we need to potentially populate a preload, and
    # not just the cache, we have to do this every time, even if the
    # desired root template is already in the cache (because its
    # requirements might not be).
    # Three things:
    # 1. we're prioritizing the case where preload is set, because that
    #    happens during render calls, which need to be fast.
    # 2. we're assuming that the total inclusions are significantly
    #    smaller than the total template cache.
    # 3. this isn't, strictly speaking, threadsafe, so if and when we add
    #    support for finite caches, we'll need to wrap this into a lock.
    #    Or something. It's not clear how this would work if someone wants
    #    both sync and async loading at the same time.
    required_loads: set[TemplateClass] = set()

    # Note that we need a preload to construct prerender trees regardless
    # of whether or not the caller cares about it.
    if preload is None:
        preload = {}

    # Provision this in advance (with mutable args) so that we can update the
    # target resource if we fetch it from the cache
    finalizer = SignatureFinalizer(
        preload=preload,
        required_loads=required_loads)

    # This is cleaner than a difference, since we need to handle both the
    # excluded and included cases if we have a preload
    for included_template_cls in signature.total_inclusions:
        if force_reload:
            required_loads.add(included_template_cls)
        elif (
            from_cache := parse_cache.get(included_template_cls)
        ) is not None:
            preload[included_template_cls] = from_cache
            if included_template_cls is template_cls:
                finalizer.target_resource = from_cache
        else:
            required_loads.add(included_template_cls)

    yield finalizer

    # Note: the finalizers need to be done separately, because the prerender
    # tree finalizer requires all of the **resources** of all inclusions to be
    # loaded prior to constructing the tree
    for required_template_cls in required_loads:
        requirement_signature = cast(
            type[TemplateIntersectable], required_template_cls
        )._templatey_signature

        # Note that we have to do this on all of the requirements;
        # otherwise, they'll be stored in the cache with an incomplete
        # signature.
        ensure_slot_tree(requirement_signature, required_template_cls)
        ensure_prerender_tree(requirement_signature, preload)

        parsed_resource = preload[required_template_cls]
        if required_template_cls is template_cls:
            finalizer.target_resource = parsed_resource

        # Note that we only want to actually cache the resources if everything
        # succeeded without error; otherwise, we could have partially-finalized
        # signatures
        parse_cache[required_template_cls] = parsed_resource

    if not hasattr(finalizer, 'target_resource'):
        raise RuntimeError(
            'Impossible branch: target template missing from load results',
            template_cls)


@dataclass(slots=True)
class _TotalityRecursionGuard:
    root_template_cls: TemplateClass
    root_total_inclusions: set[TemplateClass]


def ensure_recursive_totality(
        signature: TemplateSignature,
        template_cls: TemplateClass,
        *,
        _recursion_guard: _TotalityRecursionGuard | None = None
        ) -> None:
    """This function constructs and populates the
    ``signature.fieldset`` and ``signature.total_inclusions``
    attributes if (and only if) either one of them is missing, and
    then recursively does the same for all non-dynamic nested slot
    classes.

    This is helpful for several reasons:
    ++  it makes sure that we have a full fieldset and descendant
        classes for the entire descendancy tree, so that we can
        construct a slot tree (and thereafter a prerender tree) for
        the class
    ++  if the attributes already exist, it does nothing, so we
        don't waste work on nested template classes
    ++  it allows parsed template resources to fall out of cache
        without discarding the work we did processing the signature
        itself

    This is meant to be called during loading (instead of, for example,
    during ``@template`` decoration time) because it maximizes the
    chances that all of the type hints are available (ie, no longer
    forward refs).

    Note that our approach here, though not computationally ideal, is
    much simpler than a more sophisticated/efficient approach. Instead
    of trying to construct all of the totalities at the same time, we
    simply construct them for the root class, and then proceed on to
    each one of its inclusions iteratively. Again, there's a lot of
    duplicate work here, but it makes it **much** easier to resolve
    recursion loops.

    This returns the total inclusions for the passed class, whether or
    not it was the root template class.
    """
    if not hasattr(signature, 'fieldset'):
        signature.fieldset = NormalizedFieldset.from_template_cls(
            template_cls)

    # If we've already calculated total inclusions, AND this is a recursive
    # call, we still need to update the caller with the total inclusions!
    # This is critical; the point of recursive calls is not to populate the
    # value on the downstream child, but rather to extract inclusions on
    # behalf of the upstream parent.
    if hasattr(signature, 'total_inclusions'):
        if _recursion_guard is not None:
            _recursion_guard.root_total_inclusions.update(
                signature.total_inclusions)

    # But if we haven't calculated total inclusions yet, obviously we still
    # need to do that, regardless of whether or not this is a recursive call.
    else:
        total_inclusions: set[TemplateClass]
        if _recursion_guard is None:
            total_inclusions = {template_cls}
            recursion_guard = _TotalityRecursionGuard(
                root_template_cls=template_cls,
                root_total_inclusions=total_inclusions)
        else:
            total_inclusions = _recursion_guard.root_total_inclusions
            # Note: this MUST be before the recursive call, or it won't protect
            # against recursion!
            total_inclusions.add(template_cls)
            recursion_guard = _recursion_guard

        # Note that in addition to saving us work (both by deduping
        # and by reusing the results from creating the fieldset),
        # this also flattens unions, aliases, etc.
        direct_inclusions: set[TemplateClass] = set()
        for _, nested_slot_cls in signature.fieldset.slotpaths:
            # Note that these do NOT include dynamic slot classes!
            direct_inclusions.add(nested_slot_cls)

        # Don't forget that templates can include themselves as a slot!
        # (Also, the total_classes already includes it).
        # This is another protection against infinite recursion.
        direct_inclusions.discard(template_cls)
        for nested_slot_cls in direct_inclusions:
            if nested_slot_cls not in total_inclusions:
                nested_signature = cast(
                    type[TemplateIntersectable], nested_slot_cls
                )._templatey_signature
                ensure_recursive_totality(
                    nested_signature,
                    nested_slot_cls,
                    _recursion_guard=recursion_guard)

        # We only want to set the value if we're being called on the root;
        # nothing else is definitely correct!
        if _recursion_guard is None:
            # Note that this needs to be before the recursive call to
            # prevent infinite recursion
            signature.total_inclusions = frozenset(total_inclusions)

            # But we do actually need to ensure recursive totality, so then
            # we need to follow up with each and every one of the inclusions.
            for nested_inclusion in total_inclusions:
                nested_signature = cast(
                    type[TemplateIntersectable], nested_inclusion
                )._templatey_signature
                # Note the difference: no _root_template_cls!
                ensure_recursive_totality(nested_signature, nested_inclusion)


@dataclass(slots=True)
class _TotalityFrame:
    """When calculating totality, we need to resolve recursion loops
    (again). ``_TotalityFrame`` objects maintain the state we need to
    use a stack-based approach for that instead of the naive recursion
    option, which ends up in infinite recursion when you get to
    nontrivial recursion loops.

    TODO: this needs a better description, that's less sloppy and
    confusing. Point is, we use this to calculate recursive totality.
    """
    slot_cls: TemplateClass
    # Note: this gets mutated during processing
    remaining_direct_inclusions: set[TemplateClass]
    signature: TemplateSignature
    total_classes: set[TemplateClass] = field(init=False)
    recursion_sources: list[tuple[_TotalityFrame, ...]] = field(
        default_factory=list)

    def __post_init__(self):
        self.total_classes = {self.slot_cls}

    @classmethod
    def from_slot_cls(cls, slot_cls: TemplateClass) -> _TotalityFrame:
        signature = cast(
            type[TemplateIntersectable], slot_cls)._templatey_signature
        # Note that in addition to saving us work (both by deduping
        # and by reusing the results from creating the fieldset),
        # this also flattens unions, aliases, etc.
        direct_inclusions: set[TemplateClass] = set()
        for _, nested_slot_cls in signature.fieldset.slotpaths:
            # Note that these do NOT include dynamic slot classes!
            direct_inclusions.add(nested_slot_cls)

        # Don't forget that templates can include themselves as a slot!
        # (Also, the total_classes already includes it).
        # This prevents infinite recursion.
        direct_inclusions.discard(slot_cls)

        return cls(
            slot_cls=slot_cls,
            remaining_direct_inclusions=direct_inclusions,
            signature=signature)

    @property
    def exhausted(self) -> bool:
        return not bool(self.remaining_direct_inclusions)

    def extract_recursion_loop(
            self,
            stack: list[_TotalityFrame],
            recursion_target: _TotalityFrame
            ) -> tuple[_TotalityFrame, ...]:
        """Given the current stack, and an up-stack recursion target,
        extracts out just the frames that are part of the recursion
        loop.
        """
        target_frames: list[_TotalityFrame] = []
        target_encountered = False
        for frame in stack:
            if frame is recursion_target:
                target_encountered = True
            elif target_encountered:
                target_frames.append(frame)

        if not target_encountered:
            raise ValueError(
                'Recursion target not found in stack!',
                stack, recursion_target)

        return tuple(target_frames)


def ensure_slot_tree(
        signature: TemplateSignature,
        template_cls: TemplateClass,):
    """After ensuring recursive totality, call this to make sure
    that the slot tree is defined on the signature.

    This is meant to be called during loading (instead of, for example,
    during ``@template`` decoration time) because it maximizes the
    chances that all of the type hints are available (ie, no longer
    forward refs).
    """
    if not hasattr(signature, 'slot_tree'):
        signature.slot_tree = build_slot_tree(template_cls)


def ensure_prerender_tree(
        signature: TemplateSignature,
        preload: dict[TemplateClass, ParsedTemplateResource]):
    """After fully loading the underlying template and all of its
    inclusions, call this to make sure that the prerender tree is
    defined on the signature.

    This is meant to be called during loading (instead of, for example,
    during ``@template`` decoration time) because it maximizes the
    chances that all of the type hints are available (ie, no longer
    forward refs).
    """
    if not hasattr(signature, 'prerender_tree'):
        signature.prerender_tree = signature.slot_tree.distill_prerender_tree(
            preload)



"""Deprecated. Use prebaked.configs instead.
"""
from __future__ import annotations

from typing import Annotated

from docnote import ClcNote
from docnote import DocnoteConfig

from templatey.interpolators import NamedInterpolator
from templatey.prebaked.configs import html_escaper
from templatey.prebaked.configs import html_verifier
from templatey.prebaked.configs import noop_escaper
from templatey.prebaked.configs import noop_verifier
from templatey.templates import TemplateConfig

# This whole thing is deprecated; use prebaked.configs instead!
DOCNOTE_CONFIG = DocnoteConfig(include_in_docs=False)


html: Annotated[
    TemplateConfig,
    ClcNote(
        '''This prebaked template config uses curly braces as the interpolator
        along with a dedicated HTML escaper and verifier, with specific
        allowlisted HTML tags for context.

        Use this if you need to write HTML templates that don't inline
        javascript, CSS, etc.
        ''')
] = TemplateConfig(
    interpolator=NamedInterpolator.CURLY_BRACES,
    variable_escaper=html_escaper,
    content_verifier=html_verifier)


html_unicon: Annotated[
    TemplateConfig,
    ClcNote(
        '''This prebaked template config uses unicode control characters as the
        interpolator along with a dedicated HTML escaper and verifier, with
        specific allowlisted HTML tags for context.

        Use this if you need to write HTML templates that make use of curly
        braces within the literal template definition -- for example, if
        your template text contains inline javascript, CSS, etc.
        ''')
] = TemplateConfig(
    interpolator=NamedInterpolator.UNICODE_CONTROL,
    variable_escaper=html_escaper,
    content_verifier=html_verifier)


trusted: Annotated[
    TemplateConfig,
    ClcNote(
        '''This prebaked template config uses curly brackets as the
        interpolator, but includes **no escaping or verification**.

        Use this:
        ++  if, and **only if**, you trust all variables and content passed
            to the template
        ++  if you don't need to use curly braces within the template itself

        One example use case would be using a template for plaintext.
        ''')
] = TemplateConfig(
    interpolator=NamedInterpolator.CURLY_BRACES,
    variable_escaper=noop_escaper,
    content_verifier=noop_verifier)


trusted_unicon: Annotated[
    TemplateConfig,
    ClcNote(
        '''This prebaked template config uses unicode control characters as the
        interpolator, but includes **no escaping or verification**.

        Use this:
        ++  if, and **only if**, you trust all variables and content passed
            to the template
        ++  if you need to use curly braces within the template itself

        One example use case would be using a template as a CSS preprocessor.
        ''')
] = TemplateConfig(
    interpolator=NamedInterpolator.UNICODE_CONTROL,
    variable_escaper=noop_escaper,
    content_verifier=noop_verifier)
